# coding=utf-8
import warnings
from numbers import Number

import numpy as np
from phi import math, struct
from phi.geom import AABox
from phi.geom.geometry import assert_same_rank
from phi.struct.tensorop import collapse

from ..domain import Domain
from .field import (Field, IncompatibleFieldTypes, StaggeredSamplePoints,
                    broadcast_at, propagate_flags_children,
                    propagate_flags_operation, propagate_flags_resample)
from .grid import CenteredGrid

_SUBSCRIPTS = ['x', 'y', 'z', 'w']


def _subname(name, i):
    if i < 4:
        return '%s.%s' % (name, _SUBSCRIPTS[i])
    else:
        return '%s.%d' % (name, i)


def _res(tensor, axis):
    if isinstance(tensor, CenteredGrid):
        tensor = tensor.data
    res = list(math.staticshape(tensor)[1:-1])
    res[axis] -= 1
    return tuple(res)


def unstack_staggered_tensor(tensor):
    tensors = math.unstack(tensor, -1)
    result = []
    for i, dim in enumerate(math.spatial_dimensions(tensor)):
        slices = [slice(None, -1) if d != dim else slice(None) for d in math.spatial_dimensions(tensor)]
        result.append(math.expand_dims(tensors[i][tuple([slice(None)] + slices)], -1))
    return result


def stack_staggered_components(tensors):
    for i, tensor in enumerate(tensors):
        paddings = [[0, 1] if d != i else [0, 0] for d in range(len(tensors))]
        tensors[i] = math.pad(tensor, [[0, 0]] + paddings + [[0, 0]])
    return math.concat(tensors, -1)


def staggered_component_box(resolution, axis, box_like=None):
    staggered_box = AABox(0, resolution) if box_like is None else AABox.to_box(box_like, resolution_hint=resolution)
    unit = np.array([(staggered_box.size[axis] / resolution[axis]) if d == axis else 0 for d in range(len(resolution))])
    box = AABox(staggered_box.lower - unit / 2, staggered_box.upper + unit / 2)
    return box


@struct.definition()
class StaggeredGrid(Field):

    def __init__(self, data, box=None, name=None, **kwargs):
        Field.__init__(self, **struct.kwargs(locals()))

    @staticmethod
    def sample(value, domain, batch_size=None, name=None):
        """
        Sampmles the value to a staggered grid.

        :param value: Either constant, staggered tensor, or Field
        :param domain: the domain defines resolution and physical size of the grid
        :type domain: Domain
        :param batch_size: batch size
        :param name: field name
        :return: Sampled values in staggered grid form matching domain resolution
        :rtype: StaggeredGrid
        """
        return domain.staggered_grid(value, batch_size=batch_size, name=name)

    @struct.variable(dependencies=[Field.name, Field.flags])
    def data(self, data):
        assert data is not None
        if math.is_tensor(data) is True:
            components = unstack_staggered_tensor(data)
        else:
            components = data
        result = []
        for cmp_idx, grid in enumerate(components):
            result.append(self._component_grid(grid, cmp_idx))
        return tuple(result)

    def _component_grid(self, grid, axis):
        resolution = list(grid.resolution if isinstance(grid, CenteredGrid) else math.staticshape(grid)[1:-1])
        resolution[axis] -= 1
        box = staggered_component_box(resolution, axis, self.box)
        if isinstance(grid, CenteredGrid):
            assert grid.component_count == 1
            assert grid.rank == self.rank
            assert grid.box == box
            if grid.extrapolation != self.extrapolation:
                grid = grid.copied_with(extrapolation=self.extrapolation)
            if grid.extrapolation_value != self.extrapolation_value:
                grid = grid.copied_with(extrapolation_value=self.extrapolation_value)
        else:
            grid = CenteredGrid(grid, box=box, extrapolation=self.extrapolation, extrapolation_value=self.extrapolation_value, name=_subname(self.name, axis), batch_size=self._batch_size, flags=propagate_flags_children(self.flags, box.rank, 1))
        return grid

    @property
    def rank(self):
        return len(self.resolution)

    @struct.derived()
    def resolution(self):
        return _res(self.data[0], 0)

    @struct.constant(dependencies=Field.data)
    def box(self, box):
        box = AABox.to_box(box, resolution_hint=self.resolution)
        assert_same_rank(len(self.data), self.box, 'StaggeredGrid.data does not match box.')
        return box

    @struct.derived()
    def dx(self):
        return self.box.size / self.resolution

    @struct.constant(default='boundary')
    def extrapolation(self, extrapolation):
        if extrapolation is None:
            return 'boundary'
        assert extrapolation in ('periodic', 'constant', 'boundary') or isinstance(extrapolation, (tuple, list)), extrapolation
        return collapse(extrapolation)

    @struct.constant(default=0.0)
    def extrapolation_value(self, value):
        return collapse(value)

    def sample_at(self, points):
        return math.concat([component.sample_at(points) for component in self.data], axis=-1)

    def at(self, other_field):
        if isinstance(other_field, StaggeredGrid) and other_field.box == self.box and np.allclose(other_field.resolution, self.resolution):
            return self
        try:
            points = other_field.points
            resampled = [centeredgrid.at(other_field) for centeredgrid in self.data]
            data = math.concat([field.data for field in resampled], -1)
            return other_field.copied_with(data=data, flags=propagate_flags_resample(self, other_field.flags, other_field.rank))
        except IncompatibleFieldTypes:
            return broadcast_at(self, other_field)
        except StaggeredSamplePoints:
            return broadcast_at(self, other_field)

    def at_centers(self):
        return self.at(self.center_points)

    @property
    def component_count(self):
        return len(self.data)

    def unstack(self):
        return self.data

    @struct.derived()
    def x(self):
        return self.data[-1]

    @struct.derived()
    def y(self):
        return self.data[-2]

    @struct.derived()
    def z(self):
        return self.data[-3]

    @property
    def points(self):
        raise StaggeredSamplePoints(self)

    @property
    def center_points(self):
        return CenteredGrid.getpoints(self.box, self.resolution)

    def __repr__(self):
        if self.is_valid:
            return 'StaggeredGrid[%s, size=%s, %s]' % ('x'.join([str(r) for r in self.resolution]), self.box.size, self.unstack()[0].dtype.data)
        else:
            return struct.Struct.__repr__(self)

    def compatible(self, other_field):
        if not other_field.has_points:
            return True
        if isinstance(other_field, StaggeredGrid):
            return self.box == other_field.box and np.all(self.resolution == other_field.resolution)
        else:
            return False

    def __dataop__(self, other, linear_if_scalar, data_operator):
        if isinstance(other, StaggeredGrid):
            assert self.compatible(other), 'Fields are not compatible: %s and %s' % (self, other)
            data = [data_operator(c1, c2) for c1, c2 in zip(self.data, other.data)]
            flags = propagate_flags_operation(self.flags + other.flags, False, self.rank, self.component_count)
        elif math.ndims(other) > 0 and math.staticshape(other)[-1] > 1:
            other_components = math.unstack(math.as_tensor(other), axis=-1, keepdims=True)
            data = [data_operator(c1, c2) for c1, c2 in zip(self.data, other_components)]
            flags = propagate_flags_operation(self.flags, False, self.rank, self.component_count)
        else:
            data = [data_operator(c1, other) for c1 in self.data]
            flags = propagate_flags_operation(self.flags, linear_if_scalar, self.rank, self.component_count)
        return self.copied_with(data=np.array(data, dtype=np.object), flags=flags)

    def staggered_tensor(self):
        tensors = [c.data for c in self.data]
        return stack_staggered_components(tensors)

    def divergence(self, physical_units=True):
        components = []
        for dim, field in enumerate(self.data):
            grad = math.axis_gradient(field.data, dim)
            if physical_units:
                grad /= self.dx[dim]
            components.append(grad)
        data = math.sum(components, 0)
        return CenteredGrid(data, self.box, name='div(%s)' % self.name, batch_size=self._batch_size)

    def padded(self, widths):
        new_grids = [grid.padded(widths) for grid in self.unstack()]
        if isinstance(widths, int):
            widths = [[widths, widths]] * self.rank
        w_lower, w_upper = np.transpose(widths)
        box = AABox(self.box.lower - w_lower * self.dx, self.box.upper + w_upper * self.dx)
        return self.copied_with(data=new_grids, box=box)

    def downsample2x(self):
        data = []
        for axis in range(self.rank):
            grid = self.unstack()[axis].data
            grid = grid[tuple([slice(None, None, 2) if d - 1 == axis else slice(None) for d in range(self.rank + 2)])]  # Discard odd indices along axis
            grid = math.downsample2x(grid, axes=tuple(filter(lambda ax2: ax2 != axis, range(self.rank))))  # Interpolate values along other axes
            data.append(grid)
        return self.with_data(data)

    @staticmethod
    def gradient(scalar_field, padding_mode='replicate'):
        assert isinstance(scalar_field, CenteredGrid)
        data = scalar_field.data
        if data.shape[-1] != 1:
            raise ValueError('input must be a scalar field')
        tensors = []
        for dim in math.spatial_dimensions(data):
            upper = math.pad(data, [[0,1] if d == dim else [0,0] for d in math.all_dimensions(data)], padding_mode)
            lower = math.pad(data, [[1,0] if d == dim else [0,0] for d in math.all_dimensions(data)], padding_mode)
            tensors.append((upper - lower) / scalar_field.dx[dim - 1])
        return StaggeredGrid(tensors, scalar_field.box, name='grad(%s)' % scalar_field.name,
                             batch_size=scalar_field._batch_size)

    @staticmethod
    def from_scalar(scalar_field, axis_forces, name=None):
        warnings.warn('StaggeredGrid.from_scalar() is deprecated. Use (scalar * axis_forces).at(staggered_grid) or StaggeredField.sample(scalar * axis_forces, domain) instead.', DeprecationWarning)
        assert isinstance(scalar_field, CenteredGrid)
        assert scalar_field.component_count == 1, 'channel must be scalar but has %d components' % scalar_field.component_count
        tensors = []
        for axis in range(scalar_field.rank):
            force = axis_forces[axis] if isinstance(axis_forces, (list, tuple)) else axis_forces[...,axis]
            if isinstance(force, Number) and force == 0:
                dims = list(math.staticshape(scalar_field.data))
                dims[axis + 1] += 1
                tensors.append(math.zeros(dims, math.dtype(scalar_field.data)))
            else:
                upper = scalar_field.axis_padded(axis, 0, 1).data
                lower = scalar_field.axis_padded(axis, 1, 0).data
                tensors.append(math.mul((upper + lower) / 2, force))
        return StaggeredGrid(tensors, scalar_field.box, name=name, batch_size=scalar_field._batch_size)



import numpy as np

import dash_core_components as dcc
import dash_html_components as html
from dash.dependencies import Input, Output
from dash.exceptions import PreventUpdate

from .dash_app import DashApp


MODEL_CONTROLS = []
MODEL_ACTIONS = []


def build_model_controls(dashapp):
    assert isinstance(dashapp, DashApp)
    controls = dashapp.app.controls
    actions = dashapp.app.actions

    if len(controls) == len(actions) == 0:
        return html.Div()

    model_floats = [control for control in controls if control.type == 'float']
    model_bools = [control for control in controls if control.type == 'bool']
    model_ints = [control for control in controls if control.type == 'int']
    model_texts = [control for control in controls if control.type == 'text']

    MODEL_CONTROLS.clear()
    MODEL_CONTROLS.extend([Input(control.id, 'n_clicks') for control in actions])
    MODEL_CONTROLS.extend([Input(control.id, 'value') for control in model_floats])
    MODEL_CONTROLS.extend([Input(control.id, 'value') for control in model_ints])
    MODEL_CONTROLS.extend([Input(control.id, 'value') for control in model_bools])
    MODEL_ACTIONS.clear()
    MODEL_ACTIONS.extend([Input(action.id, 'n_clicks') for action in actions])

    layout = html.Div(style={'width': '75%', 'margin-left': 'auto', 'margin-right': 'auto', 'background-color': '#F0F0F0'}, children=[
        html.Div(id='control-div', style={'width': '95%', 'height': '90%', 'margin-left': 'auto', 'margin-right': 'auto', 'margin-top': 15, 'margin-bottom': 'auto'}, children=[
            dcc.Interval(id='initialize-controls', interval=200, max_intervals=1)
        ]),
    ])

    @dashapp.dash.callback(Output('control-div', 'children'), [Input('initialize-controls', 'n_intervals')])
    def build_controls(_):
        model_buttons = [html.Button(control.name, id=control.id) for control in actions]
        model_sliders_float = create_sliders(model_floats)
        model_sliders_int = create_sliders(model_ints)

        model_checkboxes = [dcc.Checklist(options=[{'label': control.name, 'value': control.id}],
                                          value=[control.id] if control.value else [], id=control.id)
                            for control in model_bools]

        model_textfields = []
        for control in model_texts:
            text_area = dcc.Textarea(placeholder=control.value, id=control.id, value=control.value, rows=1,
                                     style={'width': '600px', 'display': 'inline-block'})
            model_textfields.append(html.Div([control.name + '  ', text_area]))
        return [dcc.Markdown('### Model')] + model_sliders_float + model_sliders_int + model_buttons + model_textfields + model_checkboxes

    for action in actions:
        @dashapp.dash.callback(Output(action.id, 'disabled'), [Input(action.id, 'n_clicks')])
        def perform_action(n_clicks, action=action):
            if n_clicks is not None:
                dashapp.app.run_action(action)
            return False

    for control in model_floats:
        @dashapp.dash.callback(Output(control.id, 'disabled'), [Input(control.id, 'value')])
        def set_model_value(slider_value, control=control):
            use_log = False if control.type == 'int' else control.editable_value.use_log_scale
            if use_log:
                value = 10.0 ** slider_value
                if value * 0.99 <= control.range[0]:
                    value = 0.0
            else:
                value = slider_value
            control.value = value
            return False

    for control in model_ints:
        @dashapp.dash.callback(Output(control.id, 'step'), [Input(control.id, 'value')])
        def set_model_value(value, control=control):
            control.value = value
            return 1

    for control in model_bools:
        @dashapp.dash.callback(Output(control.id, 'style'), [Input(control.id, 'value')])
        def set_model_bool(values, control=control):
            control.value = True if values else False
            return {}

    for control in model_texts:
        @dashapp.dash.callback(Output(control.id, 'disabled'), [Input(control.id, 'value')])
        def set_model_text(value, control=control):
            if value is not None:
                control.value = value
            return False

    return layout


def create_sliders(controls):
    sliders = []
    for control in controls:
        val = control.value
        lower, upper = control.range
        use_log = False if control.type == 'int' else control.editable_value.use_log_scale
        if use_log:
            magn = np.log10(val)
            slider_min = np.log10(lower)
            slider_max = np.log10(upper)
            stepsize_magn = 0.1
            marks = {e: '{:.1e}'.format(np.power(10.0, e)) for e in range(-20, 20) if slider_min <= e <= slider_max}
            slider = dcc.Slider(min=slider_min, max=slider_max, value=magn, id=control.id, step=stepsize_magn, updatemode='drag', marks=marks)
        else:
            if control.type == 'int':
                marks = {v: str(v) for v in range(lower, upper + 1)}
                step = 1
            else:
                marks = {float(v): str(round(v, 4)) for v in np.linspace(lower, upper, 21)}
                step = (upper - lower) / (len(marks) - 1)
            slider = dcc.Slider(min=lower, max=upper, value=val, id=control.id, step=step, marks=marks, updatemode='drag')
        slider_container = html.Div(style={'height': '50px', 'width': '100%', 'display': 'inline-block'}, children=[control.name, slider])
        sliders.append(slider_container)
    return sliders



# coding=utf-8
from phi import math
from phi import struct
from phi.physics.domain import Domain
from phi.physics.field import CenteredGrid
from phi.physics.material import Material
from phi.struct.functions import mappable


class PoissonSolver(object):
    """
    Base class for Poisson solvers
    """

    def __init__(self, name, supported_devices, supports_guess, supports_loop_counter, supports_continuous_masks):
        """Assign details such as name, supported device (CPU/GPU), etc."""
        self.name = name
        self.supported_devices = supported_devices
        self.supports_guess = supports_guess
        self.supports_loop_counter = supports_loop_counter
        self.supports_continuous_masks = supports_continuous_masks

    def solve(self, field, domain, guess, enable_backprop):
        """
        Solves the Poisson equation Δp = d for p for all active fluid cells where active cells are given by the active_mask.
        p is expected to fulfill (Δp-d) ≤ accuracy for every active cell.

        :param enable_backprop: Whether automatic differentiation should be enabled. If False, the gradient of this operation is undefined.
        :param field: scalar input field to the solve, e.g. the divergence of the velocity channel, ∇·v
        :param domain: DomainState object specifying boundary conditions and active/fluid masks. The domain must be equal for all examples (batch dimension equal to 1).
        :param guess: (Optional) Pressure channel which can be used as an initial state for the solver
        :return: pressure tensor (same shape as divergence tensor), number of iterations (integer, 1D integer tensor or None if unknown)
        """
        raise NotImplementedError(self.__class__)

    def __repr__(self):
        """representation = name"""
        return self.name

    def __and__(self, other):
        if isinstance(self, _PoissonSolverChain):
            return _PoissonSolverChain(self.solvers + (other,))
        if isinstance(other, _PoissonSolverChain):
            return _PoissonSolverChain((self,) + other.solvers)
        else:
            return _PoissonSolverChain([self, other])


PressureSolver = PoissonSolver


@struct.definition()
class PoissonDomain(struct.Struct):

    def __init__(self, domain, valid_state=(), active=None, accessible=None, **kwargs):
        struct.Struct.__init__(self, **struct.kwargs(locals(), ignore='valid_state'))
        self._valid_state = valid_state

    @struct.constant()
    def domain(self, domain):
        assert isinstance(domain, Domain), domain
        return domain

    @struct.constant(dependencies='domain')
    def active(self, active):
        extrapolation = _active_extrapolation(Material.extrapolation_mode(self.domain.boundaries))
        if active is not None:
            assert isinstance(active, CenteredGrid)
            assert active.rank == self.domain.rank
            assert active.component_count == 1
            if active.extrapolation != extrapolation:
                active = active.copied_with(extrapolation=extrapolation)
            return active
        else:
            return self.domain.centered_grid(1, extrapolation=extrapolation)

    @struct.constant(dependencies='domain')
    def accessible(self, accessible):
        if accessible is not None:
            assert isinstance(accessible, CenteredGrid)
            assert accessible.rank == self.domain.rank
            assert accessible.component_count == 1
            return accessible
        else:
            return self.domain.centered_grid(1, extrapolation=Material.accessible_extrapolation_mode(self.domain.boundaries))

    @property
    def rank(self):
        return self.domain.rank

    def active_tensor(self, extend=0):
        """
        Scalar channel encoding active cells as ones and inactive (open/obstacle) as zero.
        Active cells are those for which physical constants_dict such as pressure or velocity are calculated.

        :param extend: Extend the grid in all directions beyond the grid size specified by the domain
        """
        return self.active.padded([[extend, extend]] * self.rank).data

    def accessible_tensor(self, extend=0):
        """
        Scalar channel encoding cells that are accessible, i.e. not solid, as ones and obstacles as zero.

        :param extend: Extend the grid in all directions beyond the grid size specified by the domain
        """
        pad_values = struct.map(lambda solid: int(not solid), Material.solid(self.domain.boundaries))
        if isinstance(pad_values, (list, tuple)):
            pad_values = [0] + list(pad_values) + [0]
        result = math.pad(self.accessible.data, [[0,0]] + [[extend, extend]] * self.rank + [[0,0]], constant_values=pad_values)
        return result

    def with_hard_boundary_conditions(self, velocity):
        masked = velocity * self._frictionless_velocity_mask(velocity)
        return masked  # TODO add surface velocity

    def _frictionless_velocity_mask(self, velocity):
        tensors = []
        for axis in range(velocity.rank):
            upper = self.accessible.padded([[0, 1] if ax == axis else [0, 0] for ax in range(self.rank)])
            lower = self.accessible.padded([[1, 0] if ax == axis else [0, 0] for ax in range(self.rank)])
            tensors.append(math.minimum(upper.data, lower.data))
        return velocity.with_data(tensors)


FluidDomain = PoissonDomain


@mappable()
def _active_extrapolation(boundaries):
    return 'periodic' if boundaries == 'periodic' else 'constant'


def poisson_solve(input_field, poisson_domain, solver=None, guess=None, gradient='implicit'):
    """
    Solves the Poisson equation Δp = input_field for p.

    :param gradient: one of ('implicit', 'autodiff', 'inverse')
        If 'autodiff', use the built-in autodiff for backpropagation.
        The intermediate results of each loop iteration will be permanently stored if backpropagation is used.
        If 'implicit', computes a forward pressure solve in reverse accumulation backpropagation.
        This requires less memory but is only accurate if the solution is fully converged.
    :param input_field: CenteredGrid
    :param poisson_domain: PoissonDomain instance
    :param solver: PoissonSolver to use, None for default
    :param guess: CenteredGrid with same size and resolution as input_field
    :return: p as CenteredGrid, iteration count as int or None if not available
    :rtype: CenteredGrid, int
    """
    assert isinstance(input_field, CenteredGrid)
    if guess is not None:
        assert isinstance(guess, CenteredGrid)
        assert guess.compatible(input_field)
        guess = guess.data
    if isinstance(poisson_domain, Domain):
        poisson_domain = PoissonDomain(poisson_domain)
    if solver is None:
        solver = _choose_solver(input_field.resolution, math.choose_backend([input_field.data, poisson_domain.active.data, poisson_domain.accessible.data]))
    if not struct.any(Material.open(poisson_domain.domain.boundaries)):  # has no open boundary
        input_field = input_field - math.mean(input_field.data, axis=tuple(range(1, 1 + input_field.rank)), keepdims=True)  # Subtract mean divergence

    assert gradient in ('autodiff', 'implicit', 'inverse')
    if gradient == 'autodiff':
        pressure, iteration = solver.solve(input_field.data, poisson_domain, guess, enable_backprop=True)
    else:
        if gradient == 'implicit':
            def poisson_gradient(_op, grad):
                return poisson_solve(CenteredGrid.sample(grad, poisson_domain.domain), poisson_domain, solver, None, gradient=gradient)[0].data
        else:  # gradient = 'inverse'
            def poisson_gradient(_op, grad):
                return CenteredGrid.sample(grad, poisson_domain.domain).laplace(physical_units=False).data
        pressure, iteration = math.with_custom_gradient(solver.solve, [input_field.data, poisson_domain, guess, False], poisson_gradient, input_index=0, output_index=0, name_base='poisson_solve')

    pressure = CenteredGrid(pressure, input_field.box, extrapolation=input_field.extrapolation, name='pressure')
    return pressure, iteration


class _PoissonSolverChain(PoissonSolver):

    def __init__(self, solvers):
        PoissonSolver.__init__(self, 'chain%s' % solvers, supported_devices=(), supports_guess=solvers[0].supports_guess, supports_loop_counter=solvers[-1].supports_loop_counter, supports_continuous_masks=False)
        self.solvers = tuple(solvers)
        for solver in solvers[1:]:
            assert isinstance(solver, PoissonSolver)
            assert solver.supports_guess

    def solve(self, field, domain, guess, enable_backprop):
        iterations = None
        for solver in self.solvers:
            guess, iterations = solver.solve(field, domain, guess, enable_backprop)
        return guess, iterations


def _choose_solver(resolution, backend):
    use_fourier = math.max(resolution) > 64
    if backend.precision == 64:
        from .fourier import FourierSolver
        from .geom import GeometricCG
        return FourierSolver() & GeometricCG(accuracy=1e-8) if use_fourier else GeometricCG(accuracy=1e-8)
    elif backend.precision == 32 and backend.matches_name('SciPy'):
        from .sparse import SparseSciPy
        return SparseSciPy()
    elif backend.precision == 32:
        from .fourier import FourierSolver
        from .sparse import SparseCG
        return FourierSolver() & SparseCG(accuracy=1e-5) if use_fourier else SparseCG(accuracy=1e-5)
    else:  # lower precision
        from .geom import GeometricCG
        return GeometricCG(accuracy=1e-2)



from phi.math.nd import downsample2x

from .stream import DerivedStream


class Downsample(DerivedStream):

    def frames(self, datasource):
        raise NotImplementedError()

    def __init__(self, field):
        DerivedStream.__init__(self, [field])
        self.field = self.inputs[0]

    def size(self, datasource, lookup=False):
        return self.field.size(datasource)

    def shape(self, datasource):
        in_shape = self.field.shape(datasource)
        return [in_shape[0]] + [s // 2 for s in in_shape[1:-1]] + [in_shape[-1]]

    def get(self, datasource, indices):
        data = self.field.get(datasource, indices)
        for array in data:
            yield downsample2x(array)



import logging
import numpy as np

from phi import math
from .solver_api import PressureSolver, FluidDomain


class MultiscaleSolver(PressureSolver):

    def __init__(self, solvers, autodiff=False):
        """
        A multigrid solver first solves the pressure on a lower-resolution grid and successively upsamples and refines it.
        On each grid, i, the pressure is calculated using the i-th provided PressureSolver.
        The resulting pressure is then upsampled and given as initial guess to the next level.

        This approach reduces the number of high-resolution iterations required, especially if the previous solver had a higher accuracy.

        :param solvers: tuple or list of PressureSolvers with length equal to number of grids
        :param autodiff: if True, use autodiff, else use multigrid forward solver for backprop
        """
        if isinstance(solvers, PressureSolver):
            solvers = [solvers] * 2
        PressureSolver.__init__(self, 'MultiscaleSolver',
                                supported_devices=solvers[0].supported_devices,
                                supports_guess=solvers[0].supports_guess,
                                supports_loop_counter=np.all([s.supports_loop_counter for s in solvers]),
                                supports_continuous_masks=True)
        assert np.all([s.supports_guess for s in solvers[1:]]), 'solvers must support initial guess'
        self.solvers = solvers
        self.autodiff = autodiff

    def solve(self, divergence, domain, pressure_guess):
        assert isinstance(domain, FluidDomain)

        if self.autodiff:
            return _mg_solve_forward(divergence, domain, pressure_guess, self.solvers)

        def pressure_gradient(op, grad):
            return _mg_solve_forward(grad, domain, None, self.solvers)[0]

        return math.with_custom_gradient(_mg_solve_forward,
                                         [divergence, domain, pressure_guess, self.solvers],
                                         pressure_gradient,
                                         input_index=0, output_index=0,
                                         name_base='multiscale_solve')


def _mg_solve_forward(divergence, domain, pressure_guess, solvers):
    fluid_mask = domain.accessible_tensor(extend=1)
    active_mask = domain.active_tensor(extend=1)
    if active_mask is not None or fluid_mask is not None:
        if not np.all([s.supports_continuous_masks for s in solvers[:-1]]):
            logging.warning(
                "MultiscaleSolver solver: There are boundary conditions inside the domain but "
                "not all intermediate solvers support continuous masks")
    div_lvls = [divergence]
    act_lvls = [active_mask]
    fld_lvls = [fluid_mask]
    for grid_i in range(len(solvers) - 1):
        div_lvls.insert(0, math.downsample2x(div_lvls[0]))
        act_lvls.insert(0, math.downsample2x(act_lvls[0]) if act_lvls[0] is not None else None)
        fld_lvls.insert(0, math.downsample2x(fld_lvls[0]) if fld_lvls[0] is not None else None)
        if pressure_guess is not None:
            pressure_guess = math.downsample2x(pressure_guess)

    iter_list = []
    for i, div in enumerate(div_lvls):
        pressure_guess, iteration = solvers[i].solve(div, FluidDomain(act_lvls[i], fld_lvls[i], boundaries), pressure_guess)
        iter_list.append(iteration)
        if pressure_guess.shape[1] < divergence.shape[1]:
            pressure_guess = math.upsample2x(pressure_guess) * 2 ** math.spatial_rank(divergence)

    return pressure_guess, iter_list



import numpy as np
from phi import math
from . import tf
from ..physics.field import Field, CenteredGrid, StaggeredGrid
from ..physics.domain import Domain
from .grid_layers import conv_layer, residual_block


def u_net(domain, input_fields, output_field, levels=2, filters=16, blocks_per_level=2, skip_combine='concat', training=False, trainable=True, reuse=None):
    """
Restrictions:
- 2D only
- Domain resolution must be multiple of 2**levels

    :param skip_combine: 'concat'
    :param blocks_per_level: number of residual blocks per level
    :param filters: Number of convolutional filters
    :type filters: int or tuple or list
    :param levels: number of additional resolution levels, equals number of downsampling / upsampling operations
    :param domain: the u-net is executed on this domain.
    :type domain: Domain
    :param input_fields: list of Fields to be passed to the network as input
    :param output_field: determines sample points of the result
    :param training: whether the network is executed in training or inference mode
    :param trainable: whether the weights of the network are trainable
    :param reuse: whether to reuse weights from previous unet calls
    :return: Field sampled like output_field
    """
    assert isinstance(domain, Domain)
    assert isinstance(output_field, Field)
    net_inputs = []
    for input_field in input_fields:
        assert isinstance(input_field, Field)
        resampled = input_field.at(domain)
        net_inputs.append(resampled)
    y = CenteredGrid.sample(math.concat([math.to_float(grid.data) for grid in net_inputs], axis=-1), domain)
    # --- Execute network ---
    pad_width = sum([2 ** i for i in range(levels)])
    y = y.padded([[0, pad_width]] * domain.rank)
    resolutions = [y]
    for level in range(levels):
        level_filters = filters if isinstance(filters, int) else filters[level]
        y = conv_layer(resolutions[0], level_filters, 2, strides=2, activation=tf.nn.relu, padding='valid', name='down_convolution_%d' % level, trainable=trainable, reuse=reuse)
        for i in range(blocks_per_level):
            y = residual_block(y, level_filters, name='down_res_block_%d_%d' % (level, i), training=training, trainable=trainable, reuse=reuse)
        resolutions.insert(0, y)

    y = resolutions.pop(0)
    assert np.all(y.box.size == domain.box.size)

    for level in range(levels):
        y = math.upsample2x(y)
        res_in = resolutions.pop(0)
        res_in = res_in.at(y)  # No resampling required, simply shaving off the top rows
        if skip_combine == 'concat':
            y = y.with_data(math.concat([y.data, res_in.data], axis=-1))
        else:
            raise NotImplementedError()
            y = y + res_in
        y = y.padded([[0, 1]] * y.rank)
        if resolutions:
            level_filters = filters if isinstance(filters, int) else reversed(filters)[level]
            y = conv_layer(y, level_filters, kernel_size=2, activation=tf.nn.relu, padding='valid', name='up_convolution_%d' % level, trainable=trainable, reuse=reuse)
            for i in range(blocks_per_level):
                y = residual_block(y, level_filters, name='up_res_block_%d_%d' % (level, i), training=training, trainable=trainable, reuse=reuse)
        else:  # Last iteration
            y = conv_layer(y, output_field.component_count, kernel_size=2, activation=None, padding='valid', name='up_convolution_%d' % level, trainable=trainable, reuse=reuse)
    result = y.at(output_field)
    return result



from phi import math
from ._geom import Geometry


def assert_same_rank(rank1, rank2, error_message):
    rank1_, rank2_ = _rank(rank1), _rank(rank2)
    if rank1_ is not None and rank2_ is not None:
        assert rank1_ == rank2_, 'Ranks do not match: %s and %s. %s' % (rank1_, rank2_, error_message)


def _rank(rank):
    if rank is None:
        return None
    elif isinstance(rank, int):
        pass
    elif isinstance(rank, Geometry):
        rank = rank.rank
    else:
        rank = math.spatial_rank(rank)
    return None if rank == 0 else rank



import numpy as np

from phi import struct, math
from ._geom import Geometry


@struct.definition()
class _NoGeometry(Geometry):

    @property
    def center(self):
        return 0

    def bounding_radius(self):
        return 0

    def bounding_half_extent(self):
        return 0

    def rank(self):
        return None

    def approximate_signed_distance(self, location):
        return math.tile(np.inf, list(math.shape(location)[:-1]) + [1])

    def lies_inside(self, location):
        return math.tile(False, list(math.shape(location)[:-1]) + [1])

    def approximate_fraction_inside(self, location, cell_size):
        return math.tile(math.to_float(0), list(math.shape(location)[:-1]) + [1])

    def shifted(self, delta):
        return self

    def rotated(self, angle):
        return self


NO_GEOMETRY = _NoGeometry()



from .field.effect import effect_applied
from .field.util import diffuse
from .physics import Physics, StateDependency


class HeatDiffusion(Physics):

    def __init__(self, diffusivity=0.1):
        Physics.__init__(self, [StateDependency('effects', 'temperature_effect', blocking=True)])
        self.diffusivity = diffusivity

    def step(self, temperature, dt=1.0, effects=()):
        # pylint: disable-msg = arguments-differ
        temperature = diffuse(temperature, dt * self.diffusivity)
        for effect in effects:
            temperature = effect_applied(effect, temperature, dt)
        return temperature.copied_with(age=temperature.age + dt)



# pylint: disable-msg = wildcard-import, unused-wildcard-import, unused-import

from phi.flow import *
from .app import *
from .session import *
from .world import *
from .data import *
from .util import *
from .grid_layers import *
from . import TF_BACKEND, tensorflow, tf

try:
    from .tf_cuda_pressuresolver import CUDASolver
except ImportError as err:
    warnings.warn("TensorFlow-CUDA solver is not available. To compile it, download phiflow sources and run\n$ python setup.py tf_cuda\nbefore reinstalling phiflow.")



import warnings
from phi import math

import tensorflow

if int(tensorflow.__version__[0]) > 1:
    warnings.warn('TensorFlow 2 is not fully supported by PhiFlow.')
    tensorflow = tensorflow.compat.v1
    tensorflow.disable_eager_execution()

tf = tensorflow

from .tf_backend import TFBackend

TF_BACKEND = TFBackend()
math.DYNAMIC_BACKEND.add_backend(TF_BACKEND)



import warnings

from .backend import Backend


class NoBackendFound(Exception):

    def __init__(self, msg):
        Exception.__init__(self, msg)


class DynamicBackend(Backend):

    def __init__(self):
        self.backends = []
        Backend.__init__(self, 'Dynamic')

    @property
    def precision(self):
        return self._precision

    @precision.setter
    def precision(self, precision):
        self._precision = precision
        for backend in self.backends:
            backend.precision = precision

    def choose_backend(self, values):
        # type: (list) -> Backend
        if not isinstance(values, tuple) and not isinstance(values, list):
            values = [values]
        for backend in self.backends:
            if backend.is_applicable(values):
                return backend
        raise NoBackendFound('No backend found for values %s; registered backends are %s' % (values, self.backends))

    def add_backend(self, backend, priority=None):
        for existing in self.backends:
            if existing.name == backend.name:
                return False
        backend.precision = self.precision
        if priority is None:
            self.backends.append(backend)
        else:
            self.backends.insert(0, backend)
        return True

    def is_applicable(self, values):
        if not isinstance(values, tuple) and not isinstance(values, list):
            values = [values]
        for backend in self.backends:
            if backend.is_applicable(values):
                return True
        return False

    def is_tensor(self, x, only_native=False):
        try:
            return self.choose_backend(x).is_tensor(x, only_native=only_native)
        except NoBackendFound:
            return False

    def as_tensor(self, x, convert_external=True):
        return self.choose_backend(x).as_tensor(x, convert_external=convert_external)

    def copy(self, tensor, only_mutable=False):
        return self.choose_backend(tensor).copy(tensor, only_mutable=only_mutable)

    def equal(self, x, y):
        return self.choose_backend([x, y]).equal(x, y)

    def random_uniform(self, shape, low=0, high=1):
        return self.choose_backend(shape).random_uniform(shape, low=low, high=high)

    def random_normal(self, shape):
        return self.choose_backend(shape).random_normal(shape)

    def stack(self, values, axis=0):
        return self.choose_backend(values).stack(values, axis)

    def concat(self, values, axis):
        return self.choose_backend(values).concat(values, axis)

    def tile(self, value, multiples):
        return self.choose_backend(value).tile(value, multiples)

    def pad(self, value, pad_width, mode='constant', constant_values=0):
        return self.choose_backend(value).pad(value, pad_width, mode, constant_values)

    def reshape(self, value, shape):
        return self.choose_backend(value).reshape(value, shape)

    def sum(self, value, axis=None, keepdims=False):
        return self.choose_backend(value).sum(value, axis=axis, keepdims=keepdims)

    def prod(self, value, axis=None):
        return self.choose_backend(value).prod(value, axis)

    def divide_no_nan(self, x, y):
        return self.choose_backend([x, y]).divide_no_nan(x, y)

    def where(self, condition, x=None, y=None):
        # For Tensorflow x,y the condition can be a Numpy array, but not the other way around. If possible, choose backend based on first input, otherwise based on condition.
        return self.choose_backend([condition, x, y]).where(condition, x, y)

    def mean(self, value, axis=None, keepdims=False):
        return self.choose_backend(value).mean(value, axis, keepdims=keepdims)

    def py_func(self, func, inputs, Tout, shape_out, stateful=True, name=None, grad=None):
        return self.choose_backend(inputs).py_func(func, inputs, Tout, shape_out, stateful, name, grad)

    def resample(self, inputs, sample_coords, interpolation='linear', boundary='constant', constant_values=0):
        return self.choose_backend([inputs, sample_coords]).resample(inputs, sample_coords, interpolation=interpolation, boundary=boundary, constant_values=constant_values)

    def range(self, start, limit=None, delta=1, dtype=None):
        return self.choose_backend([start, limit, delta]).range(start, limit, delta, dtype)

    def zeros_like(self, tensor):
        return self.choose_backend(tensor).zeros_like(tensor)

    def ones_like(self, tensor):
        return self.choose_backend(tensor).ones_like(tensor)

    def dot(self, a, b, axes):
        return self.choose_backend([a, b]).dot(a, b, axes)

    def matmul(self, A, b):
        return self.choose_backend([A, b]).matmul(A, b)

    def einsum(self, equation, *tensors):
        return self.choose_backend(tensors).einsum(equation, *tensors)

    def while_loop(self, cond, body, loop_vars, shape_invariants=None, parallel_iterations=10, back_prop=True, swap_memory=False, name=None, maximum_iterations=None):
        return self.choose_backend(loop_vars).while_loop(cond, body, loop_vars, shape_invariants, parallel_iterations, back_prop, swap_memory, name, maximum_iterations)

    def abs(self, x):
        return self.choose_backend(x).abs(x)

    def sign(self, x):
        return self.choose_backend(x).sign(x)

    def round(self, x):
        return self.choose_backend(x).round(x)

    def ceil(self, x):
        return self.choose_backend(x).ceil(x)

    def floor(self, x):
        return self.choose_backend(x).floor(x)

    def max(self, x, axis=None, keepdims=False):
        return self.choose_backend(x).max(x, axis=axis, keepdims=keepdims)

    def min(self, x, axis=None, keepdims=False):
        return self.choose_backend(x).min(x, axis=axis, keepdims=keepdims)

    def maximum(self, a, b):
        return self.choose_backend([a,b]).maximum(a, b)

    def minimum(self, a, b):
        return self.choose_backend([a,b]).minimum(a, b)

    def clip(self, x, minimum, maximum):
        return self.choose_backend([x, minimum, maximum]).clip(x, minimum, maximum)

    def with_custom_gradient(self, function, inputs, gradient, input_index=0, output_index=None, name_base='custom_gradient_func'):
        return self.choose_backend(inputs[0]).with_custom_gradient(function, inputs, gradient, input_index, output_index, name_base)

    def sqrt(self, x):
        return self.choose_backend(x).sqrt(x)

    def exp(self, x):
        return self.choose_backend(x).exp(x)

    def conv(self, tensor, kernel, padding='SAME'):
        return self.choose_backend([tensor, kernel]).conv(tensor, kernel, padding)

    def expand_dims(self, a, axis=0, number=1):
        return self.choose_backend(a).expand_dims(a, axis, number)

    def shape(self, tensor):
        return self.choose_backend(tensor).shape(tensor)

    def to_float(self, x, float64=False):
        if float64:
            warnings.warn('float64 argument is deprecated, set Backend.precision = 64 to use 64 bit operations.', DeprecationWarning)
        return self.choose_backend(x).to_float(x, float64=float64)

    def staticshape(self, tensor):
        return self.choose_backend(tensor).staticshape(tensor)

    def to_int(self, x, int64=False):
        return self.choose_backend(x).to_int(x, int64=int64)

    def to_complex(self, x):
        return self.choose_backend(x).to_complex(x)

    def gather(self, values, indices):
        return self.choose_backend([values]).gather(values, indices)

    def gather_nd(self, values, indices, batch_dims=0):
        return self.choose_backend([values]).gather_nd(values, indices, batch_dims=batch_dims)

    def unstack(self, tensor, axis=0, keepdims=False):
        return self.choose_backend(tensor).unstack(tensor, axis, keepdims=keepdims)

    def std(self, x, axis=None, keepdims=False):
        return self.choose_backend(x).std(x, axis, keepdims=keepdims)

    def boolean_mask(self, x, mask):
        return self.choose_backend([x, mask]).boolean_mask(x, mask)

    def isfinite(self, x):
        return self.choose_backend(x).isfinite(x)

    def scatter(self, points, indices, values, shape, duplicates_handling='undefined'):
        return self.choose_backend([points, indices, values]).scatter(points, indices, values, shape, duplicates_handling=duplicates_handling)

    def any(self, boolean_tensor, axis=None, keepdims=False):
        return self.choose_backend(boolean_tensor).any(boolean_tensor, axis=axis, keepdims=keepdims)

    def all(self, boolean_tensor, axis=None, keepdims=False):
        return self.choose_backend(boolean_tensor).all(boolean_tensor, axis=axis, keepdims=keepdims)

    def fft(self, x):
        return self.choose_backend(x).fft(x)

    def ifft(self, k):
        return self.choose_backend(k).ifft(k)

    def imag(self, complex):
        return self.choose_backend(complex).imag(complex)

    def real(self, complex):
        return self.choose_backend(complex).real(complex)

    def cast(self, x, dtype):
        return self.choose_backend(x).cast(x, dtype)

    def sin(self, x):
        return self.choose_backend(x).sin(x)

    def cos(self, x):
        return self.choose_backend(x).cos(x)

    def sparse_tensor(self, indices, values, shape):
        return self.choose_backend([indices, values]).sparse_tensor(indices, values, shape)

    def dtype(self, array):
        return self.choose_backend(array).dtype(array)

    def add(self, a, b):
        return self.choose_backend([a, b]).add(a, b)

    def sub(self, a, b):
        return self.choose_backend([a, b]).sub(a, b)

    def mul(self, a, b):
        return self.choose_backend([a, b]).mul(a, b)

    def div(self, numerator, denominator):
        return self.choose_backend([numerator, denominator]).div(numerator, denominator)

    def pow(self, base, exp):
        return self.choose_backend([base, exp]).pow(base, exp)

    def mod(self, dividend, divisor):
        return self.choose_backend([dividend, divisor]).mod(dividend, divisor)


DYNAMIC_BACKEND = DynamicBackend()



"""
Definition of Fluid, IncompressibleFlow as well as fluid-related functions.
"""
import warnings
from numbers import Number

import numpy as np
import six

from phi import math, struct
from phi.geom import union
from phi.physics.field import Field, mask
from phi.physics.field.angular_velocity import AngularVelocity

from .domain import Domain, DomainState
from .field import CenteredGrid, StaggeredGrid, advect, union_mask
from .field.effect import Gravity, effect_applied, gravity_tensor, FieldEffect, FieldPhysics
from .material import OPEN, Material
from .physics import Physics, StateDependency
from .pressuresolver.solver_api import FluidDomain, poisson_solve


@struct.definition()
class Fluid(DomainState):
    """
    A Fluid state consists of a density field (centered grid) and a velocity field (staggered grid).
    """

    def __init__(self, domain, density=0.0, velocity=0.0, buoyancy_factor=0.0, tags=('fluid', 'velocityfield', 'velocity'), name='fluid', **kwargs):
        DomainState.__init__(self, **struct.kwargs(locals()))

    def default_physics(self): return INCOMPRESSIBLE_FLOW

    @struct.variable(default=0, dependencies=DomainState.domain)
    def density(self, density):
        """
The marker density is stored in a CenteredGrid with dimensions matching the domain.
It describes the number of particles per physical volume.
        """
        return self.centered_grid('density', density)

    @struct.variable(default=0, dependencies=DomainState.domain)
    def velocity(self, velocity):
        """
The velocity is stored in a StaggeredGrid with dimensions matching the domain.
        """
        return self.staggered_grid('velocity', velocity)

    @struct.constant(default=0.0)
    def buoyancy_factor(self, fac):
        """
The default fluid physics can apply Boussinesq buoyancy as an upward force, proportional to the density.
This force is scaled with the buoyancy_factor (float).
        """
        return fac

    @struct.variable(default={}, holds_data=False)
    def solve_info(self, solve_info):
        return dict(solve_info)

    def __repr__(self):
        return "Fluid[density: %s, velocity: %s]" % (self.density, self.velocity)


def create_smoke(domain, density=0.0, velocity=0.0, buoyancy_factor=0.1):
    velocity_field = StaggeredGrid.sample(velocity, domain, name='velocity')
    velocity_physics = IncompressibleVFlow(domain.boundaries)
    density_field = CenteredGrid.sample(density, domain, name='density')
    density_physics = [Drift(), FieldPhysics('density')]
    buoyancy_state = FieldEffect(None, targets='velocity')
    buoyancy_physics = ProportionalGForce('density', -buoyancy_factor)
    return (velocity_field, density_field, buoyancy_state), (velocity_physics, density_physics, buoyancy_physics)


class IncompressibleFlow(Physics):
    """
Physics modelling the incompressible Navier-Stokes equations.
Supports buoyancy proportional to the marker density.
Supports obstacles, density effects, velocity effects, global gravity.
    """

    def __init__(self, pressure_solver=None, make_input_divfree=False, make_output_divfree=True, conserve_density=True):
        Physics.__init__(self, [StateDependency('obstacles', 'obstacle', blocking=True),
                                StateDependency('gravity', 'gravity', single_state=True),
                                StateDependency('density_effects', 'density_effect', blocking=True),
                                StateDependency('velocity_effects', 'velocity_effect', blocking=True)])
        self.pressure_solver = pressure_solver
        self.make_input_divfree = make_input_divfree
        self.make_output_divfree = make_output_divfree
        self.conserve_density = conserve_density

    def step(self, fluid, dt=1.0, obstacles=(), gravity=Gravity(), density_effects=(), velocity_effects=()):
        # pylint: disable-msg = arguments-differ
        gravity = gravity_tensor(gravity, fluid.rank)
        velocity = fluid.velocity
        density = fluid.density
        if self.make_input_divfree:
            velocity, solve_info = divergence_free(velocity, fluid.domain, obstacles, pressure_solver=self.pressure_solver, return_info=True)
        # --- Advection ---
        density = advect.semi_lagrangian(density, velocity, dt=dt)
        velocity = advected_velocity = advect.semi_lagrangian(velocity, velocity, dt=dt)
        if self.conserve_density and np.all(Material.solid(fluid.domain.boundaries)):
            density = density.normalized(fluid.density)
        # --- Effects ---
        for effect in density_effects:
            density = effect_applied(effect, density, dt)
        for effect in velocity_effects:
            velocity = effect_applied(effect, velocity, dt)
        velocity += (density * -gravity * fluid.buoyancy_factor * dt).at(velocity)
        divergent_velocity = velocity
        # --- Pressure solve ---
        if self.make_output_divfree:
            velocity, solve_info = divergence_free(velocity, fluid.domain, obstacles, pressure_solver=self.pressure_solver, return_info=True)
        solve_info['advected_velocity'] = advected_velocity
        solve_info['divergent_velocity'] = divergent_velocity
        return fluid.copied_with(density=density, velocity=velocity, age=fluid.age + dt, solve_info=solve_info)


class IncompressibleVFlow(Physics):

    def __init__(self, boundaries, pressure_solver=None):
        Physics.__init__(self, dependencies=[
            StateDependency('obstacles', 'obstacle'),
            StateDependency('velocity_effects', 'velocity_effect', blocking=True),
        ])
        self.boundaries = boundaries
        self.pressure_solver = pressure_solver

    def step(self, velocity, dt=1.0, obstacles=(), velocity_effects=()):
        velocity = advect.semi_lagrangian(velocity, velocity, dt=dt)
        for effect in velocity_effects:  # this is where buoyancy is applied
            velocity = effect_applied(effect, velocity, dt)
        velocity, _solve_info = divergence_free(velocity, Domain(velocity.resolution, self.boundaries, velocity.box), obstacles, pressure_solver=self.pressure_solver, return_info=True)
        return velocity.copied_with(age=velocity.age + dt)


INCOMPRESSIBLE_FLOW = IncompressibleFlow()


class Drift(Physics):
    """
Passive advection with external velocity field.
This Physics requires the world to contain a single velocity field or velocity-carrying state such as Fluid.

This Physics can be applied to all built-in Fields.
The fields will then be advected with the velocity field each time step.
    """

    def __init__(self, use_updated_velocity=False, conserve=True, velocity_field_name='velocity'):
        Physics.__init__(self, dependencies=[StateDependency('velocity', velocity_field_name, single_state=True, blocking=use_updated_velocity)])
        self.conserve = conserve

    def step(self, field, dt=1.0, velocity=None):
        if not isinstance(velocity, Field):
            velocity = velocity.velocity
        advected = advect.advect(field, velocity, dt=dt).copied_with(age=field.age + dt)
        if self.conserve and isinstance(field, (CenteredGrid, StaggeredGrid)) and np.all(~np.char.equal(struct.flatten(field.extrapolation), 'constant')):  # If field has zero extrapolation, it cannot be conserved
            advected = advected.normalized(field)
        return advected


def buoyancy(density, gravity, buoyancy_factor):
    """
Computes the buoyancy force proportional to the density.
    :param density: CenteredGrid
    :param gravity: vector or float
    :param buoyancy_factor: float
    :return: StaggeredGrid for the domain of the density
    """
    warnings.warn('buoyancy() is deprecated. Use (density * -gravity * buoyancy_factor).at(target_grid) instead.', DeprecationWarning)
    if isinstance(gravity, (int, float)):
        gravity = math.to_float(math.as_tensor([gravity] + ([0] * (density.rank - 1))))
    result = StaggeredGrid.from_scalar(density, -gravity * buoyancy_factor)
    return result


class ProportionalGForce(Physics):
    """
Computes a force field proportional to the scalar `source` field that points in the direction of gravity.
A ProportionalGForce object must be accompanied by a FieldEffect state object.
    """

    def __init__(self, source, factor):
        Physics.__init__(self, dependencies=[
            StateDependency('source_field', source, single_state=True, blocking=True),
            StateDependency('gravity', 'gravity', single_state=True)
        ])
        self.factor = factor

    def step(self, effect, dt=1.0, source_field=None, gravity=Gravity()):
        gravity = gravity_tensor(gravity, source_field.rank)
        return effect.copied_with(field=source_field * gravity * self.factor)


def _is_div_free(velocity, is_div_free):
    assert is_div_free in (True, False, None)
    if isinstance(is_div_free, bool):
        return is_div_free
    if isinstance(velocity, Number):
        return True
    return False


def solve_pressure(divergence, fluiddomain, pressure_solver=None, guess=None):
    """
Computes the pressure from the given velocity divergence using the specified solver.
    :param divergence: CenteredGrid
    :param fluiddomain: FluidDomain instance
    :param pressure_solver: PressureSolver to use, None for default
    :param guess: CenteredGrid with same size and resolution as divergence
    :return: pressure field, iteration count
    :rtype: CenteredGrid, int
    """
    return poisson_solve(divergence, fluiddomain, solver=pressure_solver, guess=guess)


def divergence_free(velocity, domain=None, obstacles=(), pressure_solver=None, return_info=False, gradient='implicit'):
    """
Projects the given velocity field by solving for and subtracting the pressure.
    :param return_info: if True, returns a dict holding information about the solve as a second object
    :param velocity: StaggeredGrid
    :param domain: Domain matching the velocity field, used for boundary conditions
    :param obstacles: list of Obstacles
    :param pressure_solver: PressureSolver. Uses default solver if none provided.
    :return: divergence-free velocity as StaggeredGrid
    """
    assert isinstance(velocity, StaggeredGrid)
    # --- Set up FluidDomain ---
    if domain is None:
        domain = Domain(velocity.resolution, OPEN)
    obstacle_mask = mask(union([obstacle.geometry for obstacle in obstacles]), antialias=False)
    if obstacle_mask is not None:
        obstacle_grid = obstacle_mask.at(velocity.center_points).copied_with(extrapolation='constant')
        active_mask = 1 - obstacle_grid
    else:
        active_mask = math.ones(domain.centered_shape(name='active', extrapolation='constant'))
    accessible_mask = active_mask.copied_with(extrapolation=Material.accessible_extrapolation_mode(domain.boundaries))
    fluiddomain = FluidDomain(domain, active=active_mask, accessible=accessible_mask)
    # --- Boundary Conditions, Pressure Solve ---
    velocity = fluiddomain.with_hard_boundary_conditions(velocity)
    for obstacle in obstacles:
        if not obstacle.is_stationary:
            obs_mask = mask(obstacle.geometry, antialias=True)
            angular_velocity = AngularVelocity(location=obstacle.geometry.center, strength=obstacle.angular_velocity, falloff=None)
            velocity = ((1 - obs_mask) * velocity + obs_mask * (angular_velocity + obstacle.velocity)).at(velocity)
    divergence_field = velocity.divergence(physical_units=False)
    pressure, iterations = poisson_solve(divergence_field, fluiddomain, solver=pressure_solver, gradient=gradient)
    pressure *= velocity.dx[0]
    gradp = StaggeredGrid.gradient(pressure)
    velocity -= fluiddomain.with_hard_boundary_conditions(gradp)
    return velocity if not return_info else (velocity, {'pressure': pressure, 'iterations': iterations, 'divergence': divergence_field})



from phi.backend.backend import Backend
from phi import math, struct
from phi.geom.geometry import assert_same_rank

from .field import Field


@struct.definition()
class AnalyticField(Field):

    def __init__(self, rank, data=None, name=None, **kwargs):
        Field.__init__(self, **struct.kwargs(locals(), ignore='rank'))
        self._rank = rank

    @property
    def rank(self):
        return self._rank

    @property
    def component_count(self):
        if self._rank is None:
            return None
        try:
            example_value = self.sample_at([[0] * self._rank])
            return math.staticshape(example_value)[-1]
        except NotImplementedError:
            return None

    def unstack(self):
        assert self.component_count is not None, 'The component_count of %s is unknown' % self
        if self.component_count == 1:
            return [self]
        else:
            components = []
            for i in range(self.component_count):
                def _context(index=i):
                    return lambda x: x.unstack()[index]
                components.append(_SymbolicOpField(_context(i), [self]))
            return components

    @property
    def points(self):
        return None

    def compatible(self, other_field):
        return True

    def __mul__(self, other):
        return _SymbolicOpField(lambda x1, x2: x1 * x2, [self, other])

    __rmul__ = __mul__

    def __div__(self, other):
        return _SymbolicOpField(lambda x1, x2: x1 / x2, [self, other])

    def __truediv__(self, other):
        return _SymbolicOpField(lambda x1, x2: x1 / x2, [self, other])

    def __sub__(self, other):
        return _SymbolicOpField(lambda x1, x2: x1 - x2, [self, other])

    def __rsub__(self, other):
        return _SymbolicOpField(lambda x1, x2: x2 - x1, [self, other])

    def __add__(self, other):
        return _SymbolicOpField(lambda x1, x2: x1 + x2, [self, other])

    __radd__ = __add__

    def __pow__(self, power, modulo=None):
        return _SymbolicOpField(lambda x1, x2: x1 ** x2, [self, power])

    def __dataop__(self, other, linear_if_scalar, data_operator):
        return _SymbolicOpField(data_operator, [self, other])

    @struct.constant(default=None)
    def data(self, data):
        return data


class SymbolicFieldBackend(Backend):
    # Abstract mehtods are overridden generically.
    # pylint: disable-msg = abstract-method

    def __init__(self, backend):
        Backend.__init__(self, 'Symbolic Field Backend')
        self.backend = backend
        for fname in dir(self):
            if fname not in ('__init__', 'is_tensor', 'is_applicable', 'symbolic_call') and not fname.startswith('__'):
                function = getattr(self, fname)
                if callable(function):
                    def context(fname=fname):
                        def proxy(*args, **kwargs):
                            return self.symbolic_call(fname, args, kwargs)
                        return proxy
                    setattr(self, fname, context())

    def symbolic_call(self, func, args, kwargs):
        assert len(kwargs) == 0, 'kwargs not supported'
        backend_func = getattr(self.backend, func)
        return _SymbolicOpField(backend_func, args)

    def is_tensor(self, x, only_native=False):
        return isinstance(x, AnalyticField)


@struct.definition()
class _SymbolicOpField(AnalyticField):

    def __init__(self, function, function_args, **kwargs):
        fields = filter(lambda arg: isinstance(arg, Field), function_args)
        AnalyticField.__init__(self, _determine_rank(fields), name=function.__name__, **struct.kwargs(locals(), ignore='fields'))
        self.fields = tuple(fields)
        self.channels = _determine_component_count(function_args)

    @struct.constant()
    def function_args(self, args):
        return args

    @struct.constant()
    def function(self, function):
        return function

    def at(self, other_field):
        args = []
        for arg in self.function_args:
            if isinstance(arg, Field):
                arg = arg.at(other_field)
            args.append(arg)
        applied = self.function(*args)
        return applied

    def sample_at(self, points):
        raise NotImplementedError()

    @property
    def component_count(self):
        return self.channels

    def unstack(self):
        unstacked = {}
        for arg in self.function_args:
            if isinstance(arg, Field):
                unstacked[arg] = arg.unstack()
            elif math.is_tensor(arg) and math.ndims(arg) > 0:
                unstacked[arg] = math.unstack(arg, axis=-1, keepdims=True)
            else:
                unstacked[arg] = [arg] * self.component_count
            assert len(unstacked[arg]) == self.component_count
        result = [_SymbolicOpField(self.function, [unstacked[arg][i] for arg in self.function_args]) for i in range(self.component_count)]
        return result

    @property
    def points(self):
        if len(self.fields) == 0:
            return None
        else:
            return self.fields[0].points

    def compatible(self, other_field):
        if len(self.fields) == 0:
            return True
        else:
            return self.fields[0].compatible(other_field)


def _determine_rank(fields):
    rank = None
    for field in fields:
        if rank is None:
            rank = field.rank
        else:
            assert_same_rank(rank, field.rank, 'All fields must have the same rank')
    return rank


def _determine_component_count(args):
    result = None
    for arg in args:
        arg_channels = None
        if isinstance(arg, Field):
            arg_channels = arg.component_count
        elif math.is_tensor(arg) and math.ndims(arg) > 0:
            arg_channels = arg.shape[-1]
        if result is None:
            result = arg_channels
        elif result == 1 and arg_channels is not None:
            result = arg_channels
        else:
            assert result == arg_channels or arg_channels is None
    return result



# coding=utf-8
import inspect
import json
import logging
import os
import os.path
import re
import shutil
import sys
import warnings

import six
import numpy as np
from os.path import join, isfile, isdir

from phi import struct, math, __version__ as phi_version
from phi.physics import field
from phi.geom import GLOBAL_AXIS_ORDER as physics_config


def read_zipped_array(filename):
    file = np.load(filename)
    array = file[file.files[-1]]  # last entry in npz file has to be data array
    if array.shape[0] != 1 or len(array.shape) == 1:
        array = np.expand_dims(array, axis=0)
    if not physics_config.is_x_first and array.shape[-1] != 1:
        array = array[..., ::-1]  # component order in stored files is always XYZ
    return array


def write_zipped_array(filename, array):
    if array.shape[0] == 1 and len(array.shape) > 1:
        array = array[0, ...]
    if not physics_config.is_x_first and array.shape[-1] != 1:
        array = array[..., ::-1]  # component order in stored files is always XYZ
    np.savez_compressed(filename, array)


def _check_same_dimensions(arrays):
    for array in arrays:
        if array.shape[1:-1] != arrays[0].shape[1:-1]:
            raise ValueError("All arrays should have the same spatial dimensions, but got %s and %s" % (array.shape, arrays[0].shape))


def read_sim_frame(simpath, fieldnames, frame, set_missing_to_none=True):
    if isinstance(fieldnames, six.string_types):
        fieldnames = [fieldnames]
    for fieldname in fieldnames:
        filename = _filename(simpath, fieldname, frame)
        if os.path.isfile(filename):
            yield read_zipped_array(filename)
        else:
            if set_missing_to_none:
                yield None
            else:
                raise IOError("Missing data at frame %d: %s" % (frame, filename))


def write_sim_frame(simpath, arrays, fieldnames, frame, check_same_dimensions=False):
    if check_same_dimensions:
        _check_same_dimensions(arrays)
    os.path.isdir(simpath) or os.mkdir(simpath)
    if not isinstance(fieldnames, (tuple, list)) and not isinstance(arrays, (tuple, list)):
        fieldnames = [fieldnames]
        arrays = [arrays]
    filenames = [_filename(simpath, name, frame) for name in fieldnames]
    for i in range(len(arrays)):
        write_zipped_array(filenames[i], arrays[i])
    return filenames


def _filename(simpath, name, frame):
    return join(simpath, "%s_%06i.npz" % (name, frame))


def read_sim_frames(simpath, fieldnames=None, frames=None):
    if fieldnames is None:
        fieldnames = get_fieldnames(simpath)
    if not fieldnames:
        return []
    if frames is None:
        frames = get_frames(simpath, fieldnames[0])
    if isinstance(frames, int):
        frames = [frames]
    single_fieldname = isinstance(fieldnames, six.string_types)
    if single_fieldname:
        fieldnames = [fieldnames]

    field_lists = [[] for f in fieldnames]
    for i in frames:
        fields = list(read_sim_frame(simpath, fieldnames, i, set_missing_to_none=False))
        for j in range(len(fieldnames)):
            field_lists[j].append(fields[j])
    result = [np.concatenate(list, 0) for list in field_lists]
    return result if not single_fieldname else result[0]


def get_fieldnames(simpath):
    fieldnames_set = {f[:-11] for f in os.listdir(simpath) if f.endswith(".npz")}
    return sorted(fieldnames_set)


def first_frame(simpath, fieldname=None):
    return min(get_frames(simpath, fieldname))


def get_frames(simpath, fieldname=None, mode="intersect"):
    if fieldname is not None:
        all_frames = {int(f[-10:-4]) for f in os.listdir(simpath) if f.startswith(fieldname) and f.endswith(".npz")}
        return sorted(all_frames)
    else:
        frames_lists = [get_frames(simpath, fieldname) for fieldname in get_fieldnames(simpath)]
        if mode.lower() == "intersect":
            intersection = set(frames_lists[0]).intersection(*frames_lists[1:])
            return sorted(intersection)
        elif mode.lower() == "union":
            if not frames_lists:
                return []
            union = set(frames_lists[0]).union(*frames_lists[1:])
            return sorted(union)


def _copy_file(source, target):
    shutil.copy(source, target)
    try:
        shutil.copystat(source, target)
    except:
        warnings.warn('Could not copy file metadata to %s' % target)


class Scene(object):

    def __init__(self, dir, category, index):
        self.dir = dir
        self.category = category
        self.index = index
        self._properties = None

    @property
    def path(self):
        return join(self.dir, self.category, "sim_%06d" % self.index)

    def subpath(self, name, create=False):
        path = join(self.path, name)
        if create and not os.path.isdir(path):
            os.mkdir(path)
        return path

    def _init_properties(self):
        if self._properties is not None:
            return
        dfile = join(self.path, "description.json")
        if isfile(dfile):
            self._properties = json.load(dfile)
        else:
            self._properties = {}

    def exists_config(self):
        return isfile(join(self.path, "description.json"))

    @property
    def properties(self):
        self._init_properties()
        return self._properties

    @properties.setter
    def properties(self, dict):
        self._properties = dict
        with open(join(self.path, "description.json"), "w") as out:
            json.dump(self._properties, out, indent=2)

    def put_property(self, key, value):
        self._init_properties()
        self._properties[key] = value
        with open(join(self.path, "description.json"), "w") as out:
            json.dump(self._properties, out, indent=2)

    def read_sim_frames(self, fieldnames=None, frames=None):
        return read_sim_frames(self.path, fieldnames=fieldnames, frames=frames)

    def read_array(self, fieldname, frame):
        return next(read_sim_frame(self.path, [fieldname], frame, set_missing_to_none=False))

    def write_sim_frame(self, arrays, fieldnames, frame, check_same_dimensions=False):
        write_sim_frame(self.path, arrays, fieldnames, frame, check_same_dimensions=check_same_dimensions)

    def write(self, obj, names=None, frame=0):
        if struct.isstruct(obj):
            obj = _transform_for_writing(obj)
            if names is None:
                names = struct.names(obj)
            values = struct.flatten(obj)
            names = struct.flatten(names)
            names = [_slugify_filename(name) for name in names]
            self.write_sim_frame(values, names, frame)
        else:
            name = str(names) if names is not None else 'unnamed'
            self.write_sim_frame([obj], [name], frame)

    def read(self, obj, frame=0):
        if struct.isstruct(obj):
            obj = _transform_for_writing(obj)
            names = struct.flatten(obj)
            if not np.all([isinstance(n, six.string_types) for n in names]):
                names = struct.names(obj)
            data = struct.map(lambda name: self.read_array(_slugify_filename(name), frame), names)
            return data
        else:
            return self.read_array('unnamed', frame)

    @property
    def fieldnames(self):
        return get_fieldnames(self.path)

    @property
    def frames(self):
        return get_frames(self.path)

    def get_frames(self, mode="intersect"):
        return get_frames(self.path, None, mode)

    def __str__(self):
        return self.path

    def __repr__(self):
        return self.path

    def copy_calling_script(self, full_trace=False, include_context_information=True):
        script_paths = [frame[1] for frame in inspect.stack()]
        script_paths = list(filter(lambda path: not _is_phi_file(path), script_paths))
        script_paths = set(script_paths) if full_trace else [script_paths[0]]
        for script_path in script_paths:
            _copy_file(script_path, join(self.subpath('src', create=True), os.path.basename(script_path)))
        if include_context_information:
            with open(join(self.subpath('src', create=True), 'context.json'), 'w') as context_file:
                json.dump({
                    'phi_version': phi_version,
                    'argv': sys.argv
                }, context_file)

    def copy_src(self, path, only_external=True):
        if not only_external or not _is_phi_file(path):
            _copy_file(path, join(self.subpath('src', create=True), os.path.basename(path)))

    def mkdir(self, subdir=None):
        path = self.path
        isdir(path) or os.mkdir(path)
        if subdir is not None:
            subpath = join(path, subdir)
            isdir(subpath) or os.mkdir(subpath)

    def remove(self):
        if isdir(self.path):
            shutil.rmtree(self.path)

    def data_paths(self, frames, field_names):
        for frame in frames:
            yield tuple([_filename(self.path, name, frame) for name in field_names])

    @staticmethod
    def create(directory, category=None, count=1, mkdir=True, copy_calling_script=True):
        if count > 1:
            scenes = []
            for _ in range(count):
                scenes.append(Scene.create(directory, category, 1, mkdir, copy_calling_script))
            return SceneBatch(scenes)
        # Single scene
        directory = os.path.expanduser(directory)
        if category is None:
            category = os.path.basename(directory)
            directory = os.path.dirname(directory)
        else:
            category = slugify(category)

        scenedir = join(directory, category)
        if not isdir(scenedir):
            os.makedirs(scenedir)
            next_index = 0
        else:
            indices = [int(name[4:]) for name in os.listdir(scenedir) if name.startswith("sim_")]
            if not indices:
                next_index = 0
            else:
                next_index = max(indices) + 1
        scene = Scene(directory, category, next_index)
        if mkdir:
            scene.mkdir()
        if copy_calling_script:
            try:
                assert mkdir
                scene.copy_calling_script()
            except IOError as err:
                warnings.warn('Failed to copy calling script to scene during Scene.create().\nCause: %s' % err)
        return scene

    @staticmethod
    def list(directory, category=None, indexfilter=None, max_count=None):
        directory = os.path.expanduser(directory)
        if not category:
            root_path = directory
            category = os.path.basename(directory)
            directory = os.path.dirname(directory)
        else:
            root_path = join(directory, category)
        if not os.path.isdir(root_path):
            return []
        indices = [int(sim[4:]) for sim in os.listdir(root_path) if sim.startswith("sim_")]
        if indexfilter:
            indices = [i for i in indices if indexfilter(i)]
        if max_count and len(indices) >= max_count:
            indices = indices[0:max_count]
        indices = sorted(indices)
        if len(indices) == 0:
            logging.warning("No simulations sim_XXXXXX found in '%s'" % root_path)
        return [Scene(directory, category, scene_index) for scene_index in indices]

    @staticmethod
    def at(sim_dir):
        sim_dir = os.path.expanduser(sim_dir)
        if sim_dir[-1] == '/':  # remove trailing backslash
            sim_dir = sim_dir[0:-1]
        dirname = os.path.basename(sim_dir)
        if not dirname.startswith("sim_"):
            raise ValueError("%s with dir %s is not a valid scene directory." % (sim_dir,dirname))
        category_directory = os.path.dirname(sim_dir)
        category = os.path.basename(category_directory)
        directory = os.path.dirname(category_directory)
        index = int(dirname[4:])
        return Scene(directory, category, index)


class SceneBatch(Scene):

    def __init__(self, scenes):
        Scene.__init__(self, scenes[0].dir, scenes[0].category, scenes[0].index)
        self.scenes = scenes

    @property
    def batch_size(self):
        return len(self.scenes)

    def write_sim_frame(self, arrays, fieldnames, frame, check_same_dimensions=False):
        for array in arrays:
            assert array.shape[0] == self.batch_size or array.shape[0] == 1,\
                'Wrong batch size: %d but %d scenes' % (array.shape[0], self.batch_size)
        for i, scene in enumerate(self.scenes):
            array_slices = [(array[i, ...] if array.shape[0] > 1 else array[0, ...]) for array in arrays]
            scene.write_sim_frame(array_slices, fieldnames, frame=frame, check_same_dimensions=check_same_dimensions)

    def read_sim_frames(self, fieldnames=None, frames=None):
        raise NotImplementedError()

    def read_array(self, fieldname, frame):
        return np.concatenate([scene.read_array(fieldname, frame) for scene in self.scenes])


def _transform_for_writing(obj):
    def f(value):
        if isinstance(value, field.StaggeredGrid):
            return value.staggered_tensor()
        if isinstance(value, field.CenteredGrid):
            return value.data
        else:
            return value
    data = struct.map(f, obj, lambda x: isinstance(x, (field.StaggeredGrid, field.CenteredGrid)), content_type='format')
    return data


def _writing_staticshape(obj):
    def f(value):
        if isinstance(value, field.StaggeredGrid):
            shape = math.staticshape(value.staggered_tensor())
            return (value._batch_size,) + shape[1:]
        if isinstance(value, field.CenteredGrid):
            return value.staticshape.data
        else:
            return math.staticshape(value)
    data = struct.map(f, obj, lambda x: isinstance(x, (field.StaggeredGrid, field.CenteredGrid)), content_type='format_staticshape')
    return data


def _slugify_filename(struct_name):
    struct_name = struct_name.replace('._', '.').replace('.', '_')
    if struct_name.startswith('_'):
        struct_name = struct_name[1:]
    return struct_name


def slugify(value):
    """
    Normalizes string, converts to lowercase, removes non-alpha characters,
    and converts spaces to hyphens.
    """
    for greek_letter, name in greek.items():
        value = value.replace(greek_letter, name)
    value = re.sub('[^\w\s-]', '', value).strip().lower()
    value = re.sub('[-\s]+', '-', value)
    return value


greek = {
    u'Α': 'Alpha',      u'α': 'alpha',
    u'Β': 'Beta',       u'β': 'beta',
    u'Γ': 'Gamma',      u'γ': 'gamma',
    u'Δ': 'Delta',      u'δ': 'delta',
    u'Ε': 'Epsilon',    u'ε': 'epsilon',
    u'Ζ': 'Zeta',       u'ζ': 'zeta',
    u'Η': 'Eta',        u'η': 'eta',
    u'Θ': 'Theta',      u'θ': 'theta',
    u'Ι': 'Iota',       u'ι': 'iota',
    u'Κ': 'Kappa',      u'κ': 'kappa',
    u'Λ': 'Lambda',     u'λ': 'lambda',
    u'Μ': 'Mu',         u'μ': 'mu',
    u'Ν': 'Nu',         u'ν': 'nu',
    u'Ξ': 'Xi',         u'ξ': 'xi',
    u'Ο': 'Omicron',    u'ο': 'omicron',
    u'Π': 'Pi',         u'π': 'pi',
    u'Ρ': 'Rho',        u'ρ': 'rho',
    u'Σ': 'Sigma',      u'σ': 'sigma',
    u'Τ': 'Tau',        u'τ': 'tau',
    u'Υ': 'Upsilon',    u'υ': 'upsilon',
    u'Φ': 'Phi',        u'φ': 'phi',
    u'Χ': 'Chi',        u'χ': 'chi',
    u'Ψ': 'Psi',        u'ψ': 'psi',
    u'Ω': 'Omega',      u'ω': 'omega',
}


def _is_phi_file(path):
    path, name = os.path.split(path)
    if name == 'phi':
        return True
    elif path == '' or name == '':
        return False
    else:
        return _is_phi_file(path)



import numpy as np
from phi import struct, math
from phi.physics.field import AnalyticField
from .grid import CenteredGrid
from .staggered_grid import StaggeredGrid
from ..domain import Domain
from ...backend.backend import Backend


@struct.definition()
class Noise(AnalyticField):
    """
Generates random noise fluctuations which can be configured in physical size and smoothness.
Each call to at() or sample_at() generates a new noise field.
Noise can be used as an initializer for CenteredGrids or StaggeredGrids.
"""

    def __init__(self, channels=1, scale=10, smoothness=1.0, math=math.DYNAMIC_BACKEND, **kwargs):
        AnalyticField.__init__(self, None, **struct.kwargs(locals()))

    @struct.constant()
    def channels(self, channels):
        """ Number of independent random scalar fields this Field consists of """
        return channels

    @struct.constant()
    def scale(self, scale):
        """ Size of noise fluctuations """
        return scale

    @struct.constant()
    def smoothness(self, smoothness):
        """ Determines how quickly high frequencies die out """
        return smoothness

    def at(self, other_field):
        if isinstance(other_field, CenteredGrid):
            batch_size = other_field._batch_size
            if batch_size is None:
                if other_field.content_type in (struct.shape, struct.staticshape):
                    batch_size = other_field.data[0]
                else:
                    batch_size = math.shape(other_field.data)[0]
            array = self.grid_sample(other_field.resolution, other_field.box.size, batch_size=batch_size)
            return other_field.with_data(array)
        if isinstance(other_field, StaggeredGrid):
            assert self.channels is None or self.channels == other_field.rank
            return other_field.with_data([self.grid_sample(grid.resolution, grid.box.size, grid._batch_size or 1, 1) for grid in other_field.unstack()])
        if isinstance(other_field, Domain):
            array = self.grid_sample(other_field.resolution, other_field.box.size)
            return CenteredGrid(array, box=other_field.box, extrapolation='boundary')

    def sample_at(self, points):
        raise NotImplementedError()

    def grid_sample(self, resolution, size, batch_size=1, channels=None):
        channels = channels or self.channels or len(size)
        shape = (batch_size,) + tuple(resolution) + (channels,)
        rndj = math.to_complex(self.math.random_normal(shape)) + 1j * math.to_complex(self.math.random_normal(shape))  # Note: there is no complex32
        k = math.fftfreq(resolution) * resolution / size * self.scale  # in physical units
        k = math.sum(k ** 2, axis=-1, keepdims=True)
        lowest_frequency = 0.1
        weight_mask = 1 / (1 + math.exp((lowest_frequency - k) * 1e3))  # High pass filter
        # --- Compute 1/k ---
        k[(0,) * len(k.shape)] = np.inf
        inv_k = 1 / k
        inv_k[(0,) * len(k.shape)] = 0
        # --- Compute result ---
        fft = rndj * inv_k ** self.smoothness * weight_mask
        array = math.real(math.ifft(fft))
        array /= math.std(array, axis=tuple(range(1, math.ndims(array))), keepdims=True)
        array -= math.mean(array, axis=tuple(range(1, math.ndims(array))), keepdims=True)
        array = math.to_float(array)
        return array

    @property
    def component_count(self):
        return self.channels

    @struct.constant()
    def math(self, m):
        assert isinstance(m, math.Backend)
        return m



from __future__ import print_function

import threading
import time

from phi.physics.field.staggered_grid import StaggeredGrid

from .stream import DerivedStream


class Interleave(DerivedStream):

    def __init__(self, fields):
        DerivedStream.__init__(self, fields)
        self.field_count = len(fields)

    def shape(self, datasource):
        return self.input_fields[0].shape(datasource)

    def size(self, datasource):
        return sum([f.size(datasource) for f in self.input_fields])

    def get(self, datasource, indices):
        for index in indices:
            yield self.input_fields[index % self.field_count].get(datasource, index // self.field_count)


class Transform(DerivedStream):

    def __init__(self, transformation, field):
        DerivedStream.__init__(self, [field])
        self.field = self.input_fields[0]
        self.transformation = transformation

    def shape(self, datasource):
        return self.field.shape(datasource)

    def size(self, datasource):
        return self.field.size(datasource)

    def get(self, datasource, indices):
        for index in indices:
            yield self.transformation(self.field.get(datasource, index))


class AsyncBatchIterator(BatchIterator):

    def __init__(self, batch_iterator, logf=None):
        # type: (BatchIterator, function) -> AsyncBatchIterator
        BatchIterator.__init__(self, batch_iterator.database, batch_iterator.dataset, batch_iterator.iterator_generator, cache=False)
        self.it = batch_iterator
        self.batches = []
        self.batch_ready = threading.Event()
        self.progressed = threading.Event()
        self.alive = True
        self.logf = logf
        threading.Thread(target=self.next_loop).start()

    def get_batch(self, streams=None, subrange=None):
        self.batch_ready.wait()
        # print("Retrieving batch %d"  % self.index)
        if not self.batches:
            raise RuntimeError("get_batch failed in AsyncBatchIterator")
        return self.batches[0]

    def progress(self):
        BatchIterator.progress(self)
        # print("Progress to batch %d" % self.index)
        del self.batches[0]
        if not self.batches:
            self.batch_ready.clear()
        self.progressed.set()

    def next_loop(self):
        while self.alive:
            while len(self.batches) < 2:
                time_before = time.time()
                self.batches.append(self.it.get_batch())
                duration = time.time() - time_before
                self.logf and self.logf("Retrieving batch %d took %f seconds." % (self.it.index, duration))
                self.batch_ready.set()
                self.it.progress()
            self.progressed.wait()
            self.progressed.clear()

    def __del__(self):
        self.alive = False



from phi import struct


class _PROPAGATOR(object):

    ALL_OPERATIONS = 'transform'
    LINEAR_OPERATIONS = 'linear'  # scaling or +/- with other field of same flag
    RESAMPLE = 'resample'
    CHILDREN = 'children'


class _FIELD_TYPE(object):

    SCALAR = 'scalar'
    VECTOR = 'vector'
    ANY = 'any'

    @staticmethod
    def list(spatial_rank, components):
        result = [_FIELD_TYPE.ANY]
        if components == 1:
            result.append(_FIELD_TYPE.SCALAR)
        if components == spatial_rank:
            result.append(_FIELD_TYPE.VECTOR)
        return result


@struct.definition()
class Flag(struct.Struct):

    def __init__(self, name, is_data_bound, is_structure_bound, propagators=(), field_types=(), **kwargs):
        struct.Struct.__init__(self, **struct.kwargs(locals()))

    @struct.constant()
    def name(self, name): return name

    @struct.constant()
    def is_data_bound(self, v):
        assert isinstance(v, bool)
        return v

    @struct.constant()
    def is_structure_bound(self, v):
        assert isinstance(v, bool)
        return v

    @struct.constant()
    def propagators(self, propagators):
        return tuple(propagators)

    @struct.constant()
    def field_types(self, field_types):
        return tuple(field_types)

    def is_applicable(self, spatial_rank, component_count):
        for type in _FIELD_TYPE.list(spatial_rank, component_count):
            if type in self.field_types:
                return True
        return False

    def propagates(self, propagator):
        return propagator in self.propagators

    def __eq__(self, other):
        return self.name == other.name

    def __hash__(self):
        return hash(self.name)

    def __repr__(self):
        return self.name


@struct.definition()
class _DivergenceFree(Flag):
    pass


DIVERGENCE_FREE = _DivergenceFree('divergence-free', True, False,
                                  propagators=[_PROPAGATOR.LINEAR_OPERATIONS, _PROPAGATOR.RESAMPLE],
                                  field_types=[_FIELD_TYPE.VECTOR])


@struct.definition()
class _L2Norm(Flag):
    pass


L2_NORMALIZED = _L2Norm('L2-normalized', True, False,
                        propagators=[],
                        field_types=[_FIELD_TYPE.ANY])


@struct.definition()
class SamplePoints(Flag):
    pass


SAMPLE_POINTS = SamplePoints('sample-points', True, True,
                             propagators=[],
                             field_types=[_FIELD_TYPE.VECTOR])



from phi.backend.backend import Backend
from . import context, struct, functions


class StructBroadcastBackend(Backend):
    # Abstract mehtods are overridden generically.
    # pylint: disable-msg = abstract-method

    def __init__(self, backend, target_content_type=struct.VALID):
        Backend.__init__(self, 'StructBroadcast')
        self.backend = backend
        self.target_content_type = target_content_type
        for fname in dir(self):
            if fname not in ('__init__', 'is_applicable', 'broadcast_function') and not fname.startswith('__'):
                function = getattr(self, fname)
                if callable(function):
                    def context(fname=fname):
                        def proxy(*args, **kwargs):
                            return self.broadcast_function(self.backend, fname, args, kwargs)
                        return proxy
                    setattr(self, fname, context())

    def is_applicable(self, values):
        for value in values:
            if struct.isstruct(value):
                return True
        return False

    def broadcast_function(self, backend, func, args, kwargs):
        backend_func = getattr(backend, func)
        obj, build_arguments = argument_assembler(args, kwargs)

        def f(*values):
            args, kwargs = build_arguments(values)
            result = backend_func(*args, **kwargs)
            return result
        target_content_type = {
            'staticshape': functions.staticshape,
            'shape': functions.shape,
            'dtype': functions.dtype,
        }.get(func, self.target_content_type)
        return functions.map(f, obj, content_type=target_content_type)


def argument_assembler(args, kwargs):
    structs, keymap = build_keymap(args, kwargs)
    assert len(structs) > 0
    if len(structs) == 1:
        obj = structs[0]
    else:
        obj = functions.zip(structs)

    def assemble_arguments(items):
        args = []
        kwargs = {}
        i = 0
        while i in keymap:
            is_struct, value = keymap[i]
            if is_struct:
                value = items[value]
            args.append(value)
            i += 1
        for key, (is_struct, value) in keymap.items():
            if not isinstance(key, int):
                if is_struct:
                    value = items[value]
                kwargs[key] = value
        return args, kwargs

    return obj, assemble_arguments


def build_keymap(args, kwargs):
    " maps key to (False, value) or (True, key); keys are integers for args and strings for kwargs "
    map = {}
    structs = []
    for i, value in enumerate(args):
        if struct.isstruct(value):
            map[i] = (True, len(structs))
            structs.append(value)
        else:
            map[i] = (False, value)
    for key, value in kwargs.items():
        if struct.isstruct(value):
            map[key] = (True, len(structs))
            structs.append(value)
        else:
            map[key] = (False, value)
    return structs, map



import warnings

from phi import struct
from phi import math


@struct.definition()
class Geometry(struct.Struct):

    @property
    def center(self):
        raise NotImplementedError()

    @property
    def rank(self):
        center = self.center
        if math.ndims(center) > 0:
            return math.staticshape(center)[-1]
        else:
            return None

    def value_at(self, location):
        """
        Samples the geometry at the given locations and returns a binary mask, labelling the points as inside=1, outside=0.

        :param location: tensor of shape (batch_size, ..., rank)
        :return: float tensor of shape (*location.shape[:-1], 1).
        """
        warnings.warn("Geometry.value_at() is deprecated. Use lies_inside or approximate_fraction_inside instead.", DeprecationWarning)
        return math.to_float(self.lies_inside(location))

    def lies_inside(self, location):
        """
        Tests whether the given location lies inside or outside of the geometry. Locations on the surface count as inside.

        :param location: float tensor of shape (batch_size, ..., rank)
        :return: bool tensor of shape (*location.shape[:-1], 1).
        """
        raise NotImplementedError(self.__class__)

    def approximate_signed_distance(self, location):
        """
        Computes the approximate distance from location to the surface of the geometry.
        Locations outside return positive values, inside negative values and zero exactly at the boundary.

        The exact distance metric used depends on the geometry.
        The approximation holds close to the surface and the distance grows to infinity as the location is moved infinitely far from the geometry.
        The distance metric is differentiable and its gradients are bounded at every point in space.

        :param location: float tensor of shape (batch_size, ..., rank)
        :return: float tensor of shape (*location.shape[:-1], 1).
        """
        raise NotImplementedError(self.__class__)

    def approximate_fraction_inside(self, other_geometry):
        """
        Computes the approximate overlap between the geometry and a small other geometry.
        Returns 1.0 if `other_geometry` is fully enclosed in this geometry and 0.0 if there is no overlap.
        Close to the surface of this geometry, the fraction filled is differentiable w.r.t. the location and size of `other_geometry`.

        To call this method on batches of geometries of same shape, pass a batched Geometry instance.
        The result tensor will match the batch shape of `other_geometry`.

        The result may only be accurate in special cases.
        The given geometries may be approximated as spheres or boxes using `bounding_radius()` and `bounding_half_extent()`.

        The default implementation of this method approximates other_geometry as a Sphere and computes the fraction using `approximate_signed_distance()`.

        :param other_geometry: (batched) Geometry instance
        :return: fraction of cell volume lying inside the geometry. float tensor of shape (other_geometry.batch_shape, 1).
        """
        assert isinstance(other_geometry, Geometry)
        radius = other_geometry.bounding_radius()
        location = other_geometry.center
        distance = self.approximate_signed_distance(location)
        inside_fraction = 0.5 - distance / radius
        inside_fraction = math.clip(inside_fraction, 0, 1)
        return inside_fraction

    def bounding_radius(self):
        """
        Returns the radius of a Sphere object that fully encloses this geometry.
        The sphere is centered at the center of this geometry.

        :return: radius of type float
        """
        raise NotImplementedError(self.__class__)

    def bounding_half_extent(self):
        """
        The bounding half-extent sets a limit on the outer-most point for each coordinate axis.
        Each component is non-negative.

        Let the bounding half-extent have value `e` in dimension `d` (`extent[...,d] = e`).
        Then, no point of the geometry lies further away from its center point than `e` along `d` (in both axis directions).

        :return: float vector
        """
        raise NotImplementedError(self.__class__)

    def shifted(self, delta):
        """
        Returns a translated version of this geometry.
        :param delta: direction vector
        :return: shifted geometry
        :rtype: Geometry
        """
        raise NotImplementedError(self.__class__)

    def rotated(self, angle):
        """
        Returns a rotated version of this geometry.
        The geometry is rotated about its center point.

        :param angle: scalar (2d) or vector (3D+) representing delta angle
        :return: rotated geometry
        :rtype: Geometry
        """
        raise NotImplementedError(self.__class__)



import os
import numpy as np

from . import tf
from phi import math
from phi.physics.pressuresolver.solver_api import PoissonSolver

# --- Load Custom Ops ---
current_dir = os.path.dirname(os.path.realpath(__file__))
kernel_path = os.path.join(current_dir, 'cuda/build/pressure_solve_op.so')
if not os.path.isfile(kernel_path):
    raise ImportError('CUDA binaries not found at %s. Run "python setup.py tf_cuda" to compile them' % kernel_path)
pressure_op = tf.load_op_library(kernel_path)


class CUDASolver(PoissonSolver):

    def __init__(self, accuracy=1e-5, max_iterations=2000):
        PoissonSolver.__init__(self, 'CUDA Conjugate Gradient', supported_devices=('GPU',), supports_loop_counter=True, supports_guess=False, supports_continuous_masks=False)
        self.accuracy = accuracy
        self.max_iterations = max_iterations

    def solve(self, divergence, domain, guess, enable_backprop):
        """
        :param guess: not used in this implementation, Kernel takes the last pressure value for initial_guess
        """
        active_mask, accessible_mask = domain.active_tensor(extend=1), domain.accessible_tensor(extend=1)
        # Setup
        dimensions = math.staticshape(divergence)[1:-1]
        dimensions = dimensions[::-1]  # the custom op needs it in the x,y,z order
        dim_array = np.array(dimensions)
        dim_product = np.prod(dimensions)
        mask_dimensions = dim_array + 2
        laplace_matrix = tf.zeros(dim_product * (len(dimensions) * 2 + 1), dtype=tf.int8)
        # Helper variables for CG, make sure new memory is allocated for each variable.
        one_vector = tf.ones(dim_product, dtype=tf.float32)
        p = tf.zeros_like(divergence, dtype=tf.float32) + 1
        z = tf.zeros_like(divergence, dtype=tf.float32) + 2
        r = tf.zeros_like(divergence, dtype=tf.float32) + 3
        pressure = tf.zeros_like(divergence, dtype=tf.float32) + 4
        # Solve
        pressure, iteration = pressure_op.pressure_solve(
            dimensions, mask_dimensions, active_mask, accessible_mask, laplace_matrix,
            divergence, p, r, z, pressure, one_vector, dim_product, self.accuracy, self.max_iterations
        )
        return pressure, iteration



import os
import warnings

from .fluidformat import Scene
from .source import DataSource, SceneSource


class Dataset(object):

    def __init__(self, name, sources=()):
        self.name = name
        self.sources = list(sources)

    def __repr__(self):
        return self.name

    def add(self, datasource):
        warnings.warn("Editing datasets is deprecated.", DeprecationWarning)
        if isinstance(datasource, DataSource):
            self.sources.append(datasource)
        else:
            for source in datasource:
                self.add(source)

    def remove(self, datasource):
        warnings.warn("Editing datasets is deprecated.", DeprecationWarning)
        self.sources.remove(datasource)

    def count(self, lookup_unknown=True):
        total = 0
        for datasource in self.sources:
            s = datasource.size(lookup=lookup_unknown)
            if s is not None:
                total += s
        return total

    def __iadd__(self, other):
        warnings.warn("Editing datasets is deprecated.", DeprecationWarning)
        if isinstance(other, DataSource):
            self.add(other)
        if isinstance(other, Dataset):
            self.sources = self.sources + other.sources
        return self

    def __add__(self, other):
        newset = Dataset("%s + %s" % (self.name, other.name))
        if isinstance(other, DataSource):
            newset.sources = self.sources + [other]
        if isinstance(other, Dataset):
            newset.sources = self.sources + other.sources
        return newset

    @staticmethod
    def load(directory, indices=None, name=None, max_scenes=None, assume_same_frames=True, assume_same_shapes=True, frames=None):
        if name is None:
            name = os.path.basename(directory)
        # --- Discover scene directories ---
        index_filter = None if indices is None else lambda i: i in indices
        scenes = Scene.list(directory, max_count=max_scenes, indexfilter=index_filter)
        sources = []
        # --- Discover frame count and  array shapes ---
        shapes = dict() if assume_same_shapes else None
        for scene in scenes:
            if assume_same_frames and frames is None:
                frames = scene.frames
            sources.append(SceneSource(scene, frames=frames, shape_map=shapes))
        # --- Create Dataset ---
        if len(sources) == 0:
            raise ValueError("No data sets found in '%s' " % directory)
        return Dataset(name, sources)



# coding=utf-8
import logging
import warnings
import numpy as np
from tensorflow.python import pywrap_tensorflow
from . import tf, TF_BACKEND

from phi import struct, math
from phi.math.math_util import is_static_shape
from phi.physics.field.staggered_grid import StaggeredGrid
from phi.physics.field.grid import CenteredGrid


def _tf_name(trace, basename):
    path = trace.path('/')
    if basename is None and len(path) == 0:
        return None
    result = path if basename is None else basename + '/' + path
    return result


def placeholder(shape, dtype=None, basename='Placeholder'):
    if struct.isstruct(dtype):
        def placeholder_map(trace):
            shape, dtype = trace.value
            return tf.placeholder(dtype, shape, _tf_name(trace, basename))
        zipped = struct.zip([shape, dtype], leaf_condition=is_static_shape)
        return struct.map(placeholder_map, zipped, leaf_condition=is_static_shape, trace=True)
    else:
        def f(trace): return tf.placeholder(TF_BACKEND.precision_dtype if dtype is None else dtype, trace.value, _tf_name(trace, basename))
        return struct.map(f, shape, leaf_condition=is_static_shape, trace=True)


def placeholder_like(obj, basename='Placeholder'):
    warnings.warn("placeholder_like may not respect the batch dimension. "
                  "For State objects, use placeholder(state.shape) instead.", DeprecationWarning, stacklevel=2)

    def f(attr): return tf.placeholder(attr.value.dtype, attr.value.shape, _tf_name(attr, basename))
    return struct.map(f, obj, leaf_condition=is_static_shape, trace=True)


def variable(initial_value, dtype=None, basename='Variable', trainable=True):
    def f(attr): return tf.Variable(attr.value, name=_tf_name(attr, basename), dtype=TF_BACKEND.precision_dtype if dtype is None else dtype, trainable=trainable)
    return struct.map(f, initial_value, trace=True)


def variable_generator(initializer, dtype=None, basename='Variable', trainable=True):
    def create_variable(shape):
        initial_value = initializer(shape)
        return variable(initial_value, dtype, basename, trainable)
    return create_variable


def constant(value, dtype=None, basename='const'):
    def f(trace): return tf.constant(trace.value, dtype=TF_BACKEND.precision_dtype if dtype is None else dtype, name=_tf_name(trace, basename))
    return struct.map(f, value, trace=True)


def is_placeholder(obj):
    return isinstance(obj, tf.Tensor) and obj.op.type == 'Placeholder'


isplaceholder = is_placeholder


def dataset_handle(shape, dtype, frames=None):
    """
Creates a single virtual TensorFlow dataset (iterator_handle) for the given struct.
The dataset is expected to hold contain all fields required for loading the obj given the current context item condition.
From the dataset, graph input tensors are derived and arranged into a struct of the same shape as obj.
If an integer is passed to frames, a list of such structs is created by unstacking the second-outer-most dimension of the dataset.
    :param shape: tensor shape or struct of tensor shapes
    :param dtype: data type of struct of data types matching shape
    :param frames: Number of frames contained in each example of the dataset. Expects shape (batch_size, frames, ...)
    :type frames: int or None
    :return: list of struct and placeholder.
     1. If frames=None: valid struct corresponding to obj. If frames>1: list thereof
     2. placeholder for a TensorFlow dataset iterator handle (dtype=string)
    :rtype: tuple
    """
    shapes = tuple(struct.flatten(shape, leaf_condition=is_static_shape))
    if struct.isstruct(dtype):
        dtypes = tuple(struct.flatten(dtype))
        assert len(dtypes) == len(shapes)
    else:
        dtypes = [dtype] * len(shapes)
    if frames is not None:
        shapes = tuple([shape[0:1] + (frames,) + shape[1:] for shape in shapes])
    # --- TF Dataset handle from string ---
    iterator_handle = tf.placeholder(tf.string, shape=[], name='dataset_iterator_handle')
    iterator = tf.data.Iterator.from_string_handle(iterator_handle, output_types=dtypes, output_shapes=shapes)
    next_element = iterator.get_next()
    # --- Create resulting struct by splitting `next_element`s ---
    if frames is None:
        next_element_list = list(next_element)
        next_struct = struct.map(lambda _: next_element_list.pop(0), shape, leaf_condition=is_static_shape)
    else:
        # --- Remap structures -> to `frames` long list of structs ---
        next_struct = []
        for frame_idx in range(frames):
            next_element_list = list(next_element)
            frame_struct = struct.map(lambda _: next_element_list.pop(0)[:, frame_idx, ...], shape, leaf_condition=is_static_shape)
            next_struct.append(frame_struct)
    return next_struct, iterator_handle


def group_normalization(x, group_count, eps=1e-5):
    batch_size, H, W, C = tf.shape(x)
    gamma = tf.Variable(np.ones([1, 1, 1, C]), dtype=TF_BACKEND.precision_dtype, name="GN_gamma")
    beta = tf.Variable(np.zeros([1, 1, 1, C]), dtype=TF_BACKEND.precision_dtype, name="GN_beta")
    x = tf.reshape(x, [batch_size, group_count, H, W, C // group_count])
    mean, var = tf.nn.moments(x, [2, 3, 4], keep_dims=True)
    x = (x - mean) / tf.sqrt(var + eps)
    x = tf.reshape(x, [batch_size, H, W, C])
    return x * gamma + beta


def residual_block(y, nb_channels, kernel_size=(3, 3), _strides=(1, 1), activation=tf.nn.leaky_relu,
                   _project_shortcut=False, padding="SYMMETRIC", name=None, training=False, trainable=True, reuse=tf.AUTO_REUSE):
    shortcut = y

    if isinstance(kernel_size, int):
        kernel_size = (kernel_size, kernel_size)

    pad1 = [(kernel_size[0] - 1) // 2, kernel_size[0] // 2]
    pad2 = [(kernel_size[1] - 1) // 2, kernel_size[1] // 2]

    # down-sampling is performed with a stride of 2
    y = tf.pad(y, [[0, 0], pad1, pad2, [0, 0]], mode=padding)
    y = tf.layers.conv2d(y, nb_channels, kernel_size=kernel_size, strides=_strides, padding='valid',
                         name=None if name is None else name + "/conv1", trainable=trainable, reuse=reuse)
    # y = tf.layers.batch_normalization(y, name=None if name is None else name+"/norm1", training=training, trainable=trainable, reuse=reuse)
    y = activation(y)

    y = tf.pad(y, [[0, 0], pad1, pad2, [0, 0]], mode=padding)
    y = tf.layers.conv2d(y, nb_channels, kernel_size=kernel_size, strides=(1, 1), padding='valid',
                         name=None if name is None else name + "/conv2", trainable=trainable, reuse=reuse)
    # y = tf.layers.batch_normalization(y, name=None if name is None else name+"/norm2", training=training, trainable=trainable, reuse=reuse)

    # identity shortcuts used directly when the input and output are of the same dimensions
    if _project_shortcut or _strides != (1, 1):
        # when the dimensions increase projection shortcut is used to match dimensions (done by 1×1 convolutions)
        # when the shortcuts go across feature maps of two sizes, they are performed with a stride of 2
        shortcut = tf.pad(shortcut, [[0, 0], pad1, pad2, [0, 0]], mode=padding)
        shortcut = tf.layers.conv2d(shortcut, nb_channels, kernel_size=(1, 1), strides=_strides, padding='valid',
                                    name=None if name is None else name + "/convid", trainable=trainable, reuse=reuse)
        # shortcut = tf.layers.batch_normalization(shortcut, name=None if name is None else name+"/normid", training=training, trainable=trainable, reuse=reuse)

    y += shortcut
    y = activation(y)

    return y


def residual_block_1d(y, nb_channels, kernel_size=(3,), _strides=(1,), activation=tf.nn.leaky_relu,
                      _project_shortcut=False, padding="SYMMETRIC", name=None, training=False, trainable=True, reuse=tf.AUTO_REUSE):
    shortcut = y

    if isinstance(kernel_size, int):
        kernel_size = (kernel_size,)

    pad1 = [(kernel_size[0] - 1) // 2, kernel_size[0] // 2]

    # down-sampling is performed with a stride of 2
    y = tf.pad(y, [[0, 0], pad1, [0, 0]], mode=padding)
    y = tf.layers.conv1d(y, nb_channels, kernel_size=kernel_size, strides=_strides, padding='valid',
                         name=None if name is None else name + "/conv1", trainable=trainable, reuse=reuse)
    # y = tf.layers.batch_normalization(y, name=None if name is None else name+"/norm1", training=training, trainable=trainable, reuse=reuse)
    y = activation(y)

    y = tf.pad(y, [[0, 0], pad1, [0, 0]], mode=padding)
    y = tf.layers.conv1d(y, nb_channels, kernel_size=kernel_size, strides=(1,), padding='valid',
                         name=None if name is None else name + "/conv2", trainable=trainable, reuse=reuse)
    # y = tf.layers.batch_normalization(y, name=None if name is None else name+"/norm2", training=training, trainable=trainable, reuse=reuse)

    # identity shortcuts used directly when the input and output are of the same dimensions
    if _project_shortcut or _strides != (1,):
        # when the dimensions increase projection shortcut is used to match dimensions (done by 1×1 convolutions)
        # when the shortcuts go across feature maps of two sizes, they are performed with a stride of 2
        shortcut = tf.pad(shortcut, [[0, 0], pad1, [0, 0]], mode=padding)
        shortcut = tf.layers.conv1d(shortcut, nb_channels, kernel_size=(1, 1), strides=_strides, padding='valid',
                                    name=None if name is None else name + "/convid", trainable=trainable, reuse=reuse)
        # shortcut = tf.layers.batch_normalization(shortcut, name=None if name is None else name+"/normid", training=training, trainable=trainable, reuse=reuse)

    y += shortcut
    y = activation(y)

    return y


def istensor(obj):
    warnings.warn("istensor is deprecated, use phi.tf.app.is_tensorflow_field instead", DeprecationWarning)
    if isinstance(obj, CenteredGrid):
        return istensor(obj.data)
    if isinstance(obj, StaggeredGrid):
        return np.any([istensor(t) for t in obj.data])
    return isinstance(obj, (tf.Tensor, tf.Variable))


def gradients(y, xs, grad_y=None):
    """
    Compute the analytic gradients using TensorFlow's automatic differentiation.

    :param y: tensor or struct of tensors. The contributions of all tensors in `y` are added up.
    :param xs: struct of input tensors
    :return: struct compatible with `xs` holding dy/dx
    """
    ys = struct.flatten(y)
    if grad_y is not None:
        grad_y = struct.flatten(grad_y)
        for i in range(len(grad_y)):
            grad_y[i] = math.cast(grad_y[i], math.dtype(ys[i]))
    xs_ = struct.flatten(xs)
    grad = tf.gradients(ys, xs_, grad_ys=grad_y)
    return struct.unflatten(grad, xs)


def stop_gradient(x):
    return struct.map(tf.stop_gradient, x)


def conv_function(scope, constants_file=None):
    if constants_file is not None:
        reader = pywrap_tensorflow.NewCheckpointReader(constants_file)

        def conv(n, filters, kernel_size, strides=[1, 1, 1, 1], padding="VALID", activation=None, name=None, kernel_initializer=None):
            assert name is not None
            kernel = reader.get_tensor("%s/%s/kernel" % (scope, name))
            assert kernel.shape[-1] == filters, "Expected %d filters but loaded kernel has shape %s for conv %s" % (kernel_size, kernel.shape, name)
            if isinstance(kernel_size, int):
                assert kernel.shape[0] == kernel.shape[1] == kernel_size
            else:
                assert kernel.shape[0:2] == kernel_size
            if isinstance(strides, int):
                strides = [1, strides, strides, 1]
            elif len(strides) == 2:
                strides = [1, strides[0], strides[1], 1]
            n = tf.nn.conv2d(n, kernel, strides=strides, padding=padding.upper(), name=name)
            if activation is not None:
                n = activation(n)
            n = tf.nn.bias_add(n, reader.get_tensor("%s/%s/bias" % (scope, name)))
            return n
    else:
        def conv(n, filters, kernel_size, strides=(1, 1), padding="valid", activation=None, name=None, kernel_initializer=None):
            with tf.variable_scope(scope):
                return tf.layers.conv2d(n, filters=filters, kernel_size=kernel_size, strides=strides, padding=padding,
                                        activation=activation, name=name, reuse=tf.AUTO_REUSE, kernel_initializer=kernel_initializer)
    return conv



import warnings

import numpy as np

import plotly.figure_factory as plotly_figures

from phi.geom import GLOBAL_AXIS_ORDER as physics_config
from phi.physics.field import CenteredGrid, StaggeredGrid
from phi.viz.plot import FRONT, RIGHT, TOP
from .colormaps import COLORMAPS

EMPTY_FIGURE = {'data': [{'z': None, 'type': 'heatmap'}]}


def dash_graph_plot(data, settings):
    # type: (object, dict) -> dict
    if data is None:
        return EMPTY_FIGURE

    if isinstance(data, np.ndarray):
        data = CenteredGrid(data)

    if isinstance(data, (CenteredGrid, StaggeredGrid)):
        component = settings.get('component', 'x')
        if data.rank == 1:
            return plot(data, settings)
        if data.rank == 2:
            if component == 'vec2' and data.component_count >= 2:
                return vector_field(data, settings)
            else:
                return heatmap(data, settings)
        if data.rank == 3:
            if component == 'vec2' and data.component_count >= 2:
                return vector_field(slice_2d(data, settings), settings)
            else:
                return heatmap(slice_2d(data, settings), settings)

    warnings.warn('No figure recipe for data %s' % data)
    return EMPTY_FIGURE


def get_color_interpolation(val, cm_arr):
    """Weighted average between point smaller and larger than it"""
    if 0 in cm_arr[:, 0] - val:
        center = cm_arr[cm_arr[:, 0] == val][-1]
    else:
        offset_positions = cm_arr[:, 0] - val
        color1 = cm_arr[np.argmax(offset_positions[offset_positions < 0])]  # largest value smaller than val
        color2 = cm_arr[np.argmin(offset_positions[offset_positions > 0])]  # smallest value larger than val
        if color1[0] == color2[0]:
            center = color1
        else:
            x = (val - color1[0]) / (color2[0] - color1[0])  # weight of row2
            center = color1 * (1 - x) + color2 * x
    center[0] = val
    return center


def get_div_map(zmin, zmax, equal_scale=False, colormap=None):
    """
    :param colormap: colormap defined as list of [fraction_val, red_frac, green_frac, blue_frac]
    :type colormap: list or array
    """
    colormap = COLORMAPS[colormap]
    # Ensure slicing
    cm_arr = np.array(colormap).astype(np.float64)
    # Centeral color
    if 0.5 not in cm_arr[:, 0]:
        central_color = get_color_interpolation(0.5, cm_arr)[1:]
    else:
        central_color = cm_arr[cm_arr[:, 0] == 0.5][-1][1:]
    # Return base
    if zmin == zmax:
        central_color = np.round(central_color).astype(np.int32)
        return [(0, "rgb({},{},{})".format(*central_color)), (1, "rgb({},{},{})".format(*central_color))]
    center = abs(zmin / (zmax - zmin))
    if zmin > 0:
        center = 0
    # Rescaling
    if not equal_scale:
        # Full range, Zero-centered
        neg_flag = cm_arr[:, 0] < 0.5
        pos_flag = cm_arr[:, 0] >= 0.5
        cm_arr[neg_flag, 0] = cm_arr[neg_flag, 0] * 2 * center  # Scale (0, 0.5) -> (0, center)
        cm_arr[pos_flag, 0] = (cm_arr[pos_flag, 0] - 0.5) * 2 * (1 - center) + center  # Scale (0.5, 1) -> (center, 0.5)
        # Drop duplicate zeros. Allow for not center value in original map.
        if zmin == 0:
            cm_arr = cm_arr[np.max(np.arange(len(cm_arr))[cm_arr[:, 0] == 0]):]
    else:
        cm_arr[:, 0] = cm_arr[:, 0] - 0.5  # center at zero (-0.5, 0.5)
        # Scale desired range
        if zmax > abs(zmin):
            cm_scale = (1 - center) / (np.max(cm_arr[:, 0]))  # scale by plositives
        else:
            cm_scale = center / (np.max(cm_arr[:, 0]))  # scale by negatives
        # Scale the maximum to +1 when centered
        cm_arr[:, 0] *= cm_scale
        cm_arr[:, 0] += center  # center
        # Add zero if it doesn't exist
        if 0 not in cm_arr[:, 0]:
            new_min = get_color_interpolation(0, cm_arr)
            cm_arr = np.vstack([new_min, cm_arr])
        # Add one if it doesn't exist
        if 1 not in cm_arr[:, 0]:
            new_max = get_color_interpolation(1, cm_arr)
            cm_arr = np.vstack([cm_arr, new_max])
        # Compare center
        #new_center = get_color_interpolation(center, cm_arr)
        # if not all(new_center == [center, *central_color]):
        #    print("Failed center comparison.")
        #    print("Center: {}".format(new_center))
        #    print("Center should be: {}".format([center, *central_color]))
        #    assert False
        # Cut to (0, 1)
        cm_arr = cm_arr[cm_arr[:, 0] >= 0]
        cm_arr = cm_arr[cm_arr[:, 0] <= 1]
    cm_str = [("{}".format(val), "rgb({:.0f},{:.0f},{:.0f})".format(*colors)) for val, colors in zip(cm_arr[:, 0], cm_arr[:, 1:])]
    return cm_str


def heatmap(data, settings):
    assert isinstance(data, (StaggeredGrid, CenteredGrid))
    assert data.rank == 2
    batch = settings.get('batch', 0)
    component = settings.get('component', 'x')

    if isinstance(data, StaggeredGrid):
        if component == 'x':
            data = data.unstack()[physics_config.x]
        elif component == 'y':
            data = data.unstack()[physics_config.y]
        elif component == 'z':
            return EMPTY_FIGURE
        elif component == 'length':
            data = data.at_centers()
        else:
            raise ValueError(component)
    z = data.data[batch, ...]
    z = reduce_component(z, component)
    if physics_config.is_x_first:
        z = np.transpose(z)
        x = data.points.data[0, :, 0, 0]
        y = data.points.data[0, 0, :, 1]
    else:
        y = data.points.data[0, :, 0, 0]
        x = data.points.data[0, 0, :, 1]
    if settings.get('slow_colorbar', False):
        z_min, z_max = settings['minmax']
    else:
        z_min, z_max = np.min(z), np.max(z)
    color_scale = get_div_map(z_min, z_max, equal_scale=True, colormap=settings.get('colormap', None))
    return {'data': [{
        'x': x,
        'y': y,
        'z': z,
        'zauto': 'false',
        'zmin': str(z_min),
        'zmax': str(z_max),
        'type': 'heatmap',
        'colorscale': color_scale,
        # 'colorbar': {'title': settings.units,  }  # TODO: Implement units into PhiFlow
    }]}


def slice_2d(field3d, settings):
    if isinstance(field3d, np.ndarray):
        field3d = CenteredGrid(field3d)
    if isinstance(field3d, StaggeredGrid):
        component = settings.get('component', 'length')
        if component in ('z', 'y', 'x'):
            field3d = field3d.unstack()[{'z': physics_config.z, 'y': physics_config.y, 'x': physics_config.x}[component] % 3]
        else:
            field3d = field3d.at_centers()
    assert isinstance(field3d, CenteredGrid) and field3d.rank == 3
    depth = settings.get('depth', 0)
    projection = settings.get('projection', FRONT)

    removed_axis = {FRONT: physics_config.y, RIGHT: physics_config.x, TOP: physics_config.z}[projection] % 3

    data = field3d.data[(slice(None),) + tuple([min(depth, field3d.resolution[i]) if i == removed_axis else slice(None) for i in range(3)]) + (slice(None),)]
    if projection == RIGHT and not physics_config.is_x_first:
        data = np.transpose(data, axes=(0, 2, 1, 3))

    return CenteredGrid(data, box=field3d.box.without_axis(removed_axis))


def plot(field1d, settings):
    assert isinstance(field1d, (CenteredGrid, StaggeredGrid))
    batch = settings.get('batch', 0)
    component = settings.get('component', 'x')
    if isinstance(field1d, StaggeredGrid):
        field1d = field1d.unstack()[0]
    assert isinstance(field1d, CenteredGrid)
    x = field1d.points.data[0, :, 0]
    data = field1d.data[min(field1d.resolution[0], batch), :, :]
    data = reduce_component(data, component)
    return {'data': [{'mode': 'markers+lines', 'type': 'scatter', 'x': x, 'y': data}]}


def reduce_component(tensor, component):
    clen = tensor.shape[-1]
    if clen == 1:
        return tensor[..., 0]
    if component == 'x':
        return tensor[..., physics_config.x]
    if component == 'y':
        return tensor[..., physics_config.y]
    if component == 'z':
        if clen >= 3:
            return tensor[..., physics_config.z]
        else:
            return np.zeros_like(tensor[..., 0])
    if component == 'length':
        return np.sqrt(np.sum(tensor**2, axis=-1, keepdims=False))
    if component == 'vec2':
        return tensor[..., (physics_config.y, physics_config.x)]


def vector_field(field2d, settings):
    assert isinstance(field2d, (CenteredGrid, StaggeredGrid))
    if isinstance(field2d, StaggeredGrid):
        field2d = field2d.at_centers()
    assert isinstance(field2d, CenteredGrid)
    assert field2d.rank == 2

    batch = settings.get('batch', 0)
    batch = min(batch, field2d.data.shape[0])

    arrow_origin = settings.get('arrow_origin', 'tip')
    assert arrow_origin in ('base', 'center', 'tip')
    max_resolution = settings.get('max_arrow_resolution', 40)
    max_arrows = settings.get('max_arrows', 2000)
    min_arrow_length = settings.get('min_arrow_length', 0.005) * np.max(field2d.box.size)
    draw_full_arrows = settings.get('draw_full_arrows', False)

    y, x = field2d.points.data[0, ..., (physics_config.y, physics_config.x)]
    data_y, data_x = field2d.data[batch, ..., (physics_config.y, physics_config.x)]

    while np.prod(x.shape) > max_resolution ** 2:
        y = y[::2, ::2]
        x = x[::2, ::2]
        data_y = data_y[::2, ::2]
        data_x = data_x[::2, ::2]

    y = y.flatten()
    x = x.flatten()
    data_y = data_y.flatten()
    data_x = data_x.flatten()

    if max_arrows is not None or min_arrow_length > 0:
        length = np.sqrt(data_y**2 + data_x**2)
        keep_indices = np.argsort(length)
        keep_indices = keep_indices[length[keep_indices] > min_arrow_length]
        if len(keep_indices) > max_arrows:
            keep_indices = keep_indices[-max_arrows:]
        y = y[keep_indices]
        x = x[keep_indices]
        data_y = data_y[keep_indices]
        data_x = data_x[keep_indices]

    if len(x) == 0:
        return EMPTY_FIGURE

    if arrow_origin == 'tip':
        x -= data_x
        y -= data_y
    elif arrow_origin == 'center':
        x -= 0.5 * data_x
        y -= 0.5 * data_y

    x_range, y_range = [field2d.box.get_lower(1), field2d.box.get_upper(1)], [field2d.box.get_lower(0), field2d.box.get_upper(0)]
    if physics_config.is_x_first:
        x_range, y_range = y_range, x_range

    if draw_full_arrows:
        result = plotly_figures.create_quiver(x, y, data_x, data_y, scale=1.0)  # 7 points per arrow
        result.update_xaxes(range=x_range)
        result.update_yaxes(range=y_range)
        return result
    else:
        lines_y = np.stack([y, y + data_y, [None] * len(x)], -1).flatten()  # 3 points per arrow
        lines_x = np.stack([x, x + data_x, [None] * len(x)], -1).flatten()
        return {
            'data': [
                {
                    'mode': 'lines',
                    'x': lines_x,
                    'y': lines_y,
                    'type': 'scatter',
                }
            ],
            'layout': {
                'xaxis': {'range': x_range},
                'yaxis': {'range': y_range},
            }
        }



import distutils.cmd
import distutils.log
import errno
import subprocess
import os
from setuptools import setup


class CudaCommand(distutils.cmd.Command):
    description = 'Compile CUDA sources'
    user_options = [
        ('gcc=', None, 'Path to the gcc compiler.'),
        ('gcc-4-8=', None, 'Path to gcc-4.8 compiler.'),
        ('nvcc=', None, 'Path to the Nvidia nvcc compiler.'),
        ('cuda-lib=', None, 'Path to the CUDA libraries.'),
    ]

    def run(self):
        src_path = os.path.abspath('./phi/tf/cuda/src')
        build_path = os.path.abspath('./phi/tf/cuda/build')
        print('Source Path:\t' + src_path)
        print('Build Path:\t' + build_path)

        # Get TF Compile/Link Flags and write to env
        import tensorflow as tf
        if tf.__version__[0] == '2':
            print('Adjusting for tensorflow 2.0')
            tf = tf.compat.v1
            tf.disable_eager_execution()
        tf_cflags = tf.sysconfig.get_compile_flags()
        tf_lflags = tf.sysconfig.get_link_flags()
        # print(tf_cflags)
        # print(tf_lflags)
        # print('lib: ' + tf.sysconfig.get_lib())

        link_cuda_lib = '-L' + self.cuda_lib

        # print(link_cuda_lib)

        # Remove old build files
        if os.path.isdir(build_path):
            print('Removing old build files from %s' % build_path)
            for file in os.listdir(build_path):
                os.remove(os.path.join(build_path, file))
        else:
            print('Creating build directory at %s' % build_path)
            os.mkdir(build_path)

        print('Compiling CUDA code...')
        # Build the Laplace Matrix Generation CUDA Kernels
        subprocess.check_call(
            [
                self.nvcc,
                '-std=c++11',
                '-c',
                '-o',
                os.path.join(build_path, 'laplace_op.cu.o'),
                os.path.join(src_path, 'laplace_op.cu.cc'),
                '-x',
                'cu',
                '-Xcompiler',
                '-fPIC'
            ]
            + tf_cflags
        )

        # Build the Laplace Matrix Generation Custom Op
        # This is only needed for the Laplace Matrix Generation Benchmark
        subprocess.check_call(
            [
                self.gcc_4_8,
                '-std=c++11',
                '-shared',
                '-o',
                os.path.join(build_path, 'laplace_op.so'),
                os.path.join(src_path, 'laplace_op.cc'),
                os.path.join(build_path, 'laplace_op.cu.o'),
                '-fPIC'
            ]
            + tf_cflags
            + tf_lflags
            + ['-L/usr/local/cuda/lib64/','-lcudart']
        )

        # Build the Pressure Solver CUDA Kernels
        subprocess.check_call(
            [
                self.nvcc,
                '-std=c++11',
                '-c',
                '-lcublas',
                '-o',
                os.path.join(build_path, 'pressure_solve_op.cu.o'),
                os.path.join(src_path, 'pressure_solve_op.cu.cc'),
                '-x', 'cu',
                '-Xcompiler',
                '-fPIC'
            ]
            + tf_cflags
        )

        # Build the Pressure Solver Custom Op
        subprocess.check_call(
            [
                self.gcc_4_8,
                '-std=c++11',
                '-shared',
                '-o',
                os.path.join(build_path, 'pressure_solve_op.so'),
                os.path.join(src_path, 'pressure_solve_op.cc'),
                os.path.join(build_path, 'pressure_solve_op.cu.o'),
                os.path.join(build_path, 'laplace_op.cu.o'),
                '-fPIC'
            ]
            + tf_cflags
            + tf_lflags
            + ['-L/usr/local/cuda/lib64/','-lcudart']
        )

        # Build the Resample CUDA Kernels
        subprocess.check_call(
            [
                self.nvcc,
                '-std=c++11',
                '-c',
                '-o',
                os.path.join(build_path, 'resample.cu.o'),
                os.path.join(src_path, 'resample.cu.cc'),
                '-D GOOGLE_CUDA=1',
                '-x', 'cu',
                '-Xcompiler',
                '-fPIC',
                '--expt-relaxed-constexpr',
                '-DNDEBUG',
                '-O3'
            ]
            + tf_cflags
        )

        # Build the Resample Custom Op
        try:
            subprocess.check_call(
                [
                    self.gcc_4_8,
                    '-std=c++11',
                    '-shared',
                    '-o',
                    os.path.join(build_path, 'resample.so'),
                    os.path.join(src_path, 'resample.cc'),
                    os.path.join(build_path, 'resample.cu.o'),
                    '-fPIC',
                    '-lcudart',
                    '-O3',
                    link_cuda_lib
                ]
                + tf_cflags
                + tf_lflags
            )
        except OSError as e:
            if e.errno == errno.ENOENT:
                print('Please install g++-4.8 as it is needed to compile the advection operator.')
                # raise e
            else:
                raise e

        # Build the Resample Gradient CUDA Kernels
        subprocess.check_call(
            [
                self.nvcc,
                '-std=c++11',
                '-c',
                '-o',
                os.path.join(build_path, 'resample_gradient.cu.o'),
                os.path.join(src_path, 'resample_gradient.cu.cc'),
                '-D GOOGLE_CUDA=1',
                '-x', 'cu',
                '-Xcompiler',
                '-fPIC',
                '--expt-relaxed-constexpr',
                '-DNDEBUG',
                '-O3'
            ]
            + tf_cflags
        )

        # Build the Resample Gradient Custom Op
        try:
            subprocess.check_call(
                [
                    self.gcc_4_8,
                    '-std=c++11',
                    '-shared',
                    '-o',
                    os.path.join(build_path, 'resample_gradient.so'),
                    os.path.join(src_path, 'resample_gradient.cc'),
                    os.path.join(build_path, 'resample_gradient.cu.o'),
                    '-fPIC',
                    '-lcudart',
                    '-O3',
                    link_cuda_lib
                ]
                + tf_cflags
                + tf_lflags
            )
        except OSError as e:
            if e.errno == errno.ENOENT:
                print('Please install g++-4.8 as it is needed to compile the advection operator.')
                # raise e
            else:
                raise e

    def initialize_options(self):
        self.gcc = 'gcc'
        self.gcc_4_8 = 'g++-4.8'
        self.nvcc = 'nvcc'
        self.cuda_lib = '/usr/local/cuda/lib64/'

    def finalize_options(self):
        assert os.path.isfile(self.gcc) or self.gcc == 'gcc'
        assert os.path.isfile(self.nvcc) or self.nvcc == 'nvcc'


try:
    with open(os.path.join(os.path.dirname(__file__), 'documentation/Package_Info.md'), 'r') as readme:
        long_description = readme.read()
except FileNotFoundError:
    pass

with open(os.path.join(os.path.dirname(__file__), 'phi', 'VERSION'), 'r') as version_file:
    version = version_file.read()


setup(
    name='phiflow',
    version=version,
    download_url='https://github.com/tum-pbs/PhiFlow/archive/%s.tar.gz' % version,
    packages=['phi',
              'phi.app',
              'phi.backend',
              'phi.data',
              'phi.geom',
              'phi.local',
              'phi.math',
              'phi.physics',
              'phi.physics.field',
              'phi.physics.pressuresolver',
              'phi.struct',
              'phi.tf',
              'phi.viz',
              'phi.viz.dash',
              'webglviewer'],
    cmdclass={
        'tf_cuda': CudaCommand,
    },
    description='Research-oriented differentiable fluid simulation framework',
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords=['Differentiable', 'Simulation', 'Fluid', 'Machine Learning', 'Deep Learning'],
    license='MIT',
    author='Philipp Holl',
    author_email='philipp.holl@tum.de',
    url='https://github.com/tum-pbs/PhiFlow',
    include_package_data=True,
    install_requires=['six', 'packaging', 'scipy'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    extras_require={
        'gui': ['dash',
                'dash-renderer',
                'dash-html-components',
                'dash-core-components',
                'plotly',
                'imageio',
                'matplotlib'],
    }
)



from phi import math, struct

from ._geom import Geometry


@struct.definition(traits=[math.BATCHED])
class Sphere(Geometry):

    def __init__(self, center, radius, **kwargs):
        Geometry.__init__(self, **struct.kwargs(locals()))

    @struct.constant(min_rank=0)
    def radius(self, radius):
        return radius

    @struct.constant(min_rank=1)
    def center(self, center):
        return center

    def lies_inside(self, location):
        center = math.batch_align(self.center, 1, location)
        radius = math.batch_align(self.radius, 0, location)
        distance_squared = math.sum((location - center) ** 2, axis=-1, keepdims=True)
        return distance_squared <= radius ** 2

    def approximate_signed_distance(self, location):
        """
Computes the exact distance from location to the closest point on the sphere.
Very close to the sphere center, the distance takes a constant value.
        :param location: float tensor of shape (batch_size, ..., rank)
        :return: float tensor of shape (*location.shape[:-1], 1).
        """
        center = math.batch_align(self.center, 1, location)
        radius = math.batch_align(self.radius, 0, location)
        distance_squared = math.sum((location - center)**2, axis=-1, keepdims=True)
        distance_squared = math.maximum(distance_squared, radius * 1e-2)  # Prevent infinite gradient at sphere center
        distance = math.sqrt(distance_squared)
        return distance - radius

    def bounding_radius(self):
        return self.radius

    def bounding_half_extent(self):
        return self.radius

    def shifted(self, delta):
        return self.copied_with(center=self.center + delta)

    def rotated(self, angle):
        return self



import collections
import numbers
import warnings

import numpy as np
import scipy.signal
import scipy.sparse

from phi.backend.backend_helper import split_multi_mode_pad, PadSettings, general_grid_sample_nd
from .backend import Backend


class SciPyBackend(Backend):

    """
    Core Python Backend using NumPy & SciPy
    """

    def __init__(self, precision=32):
        Backend.__init__(self, "SciPy", precision=precision)

    @property
    def precision_dtype(self):
        return {16: np.float16, 32: np.float32, 64: np.float64, None: np.float32}[self.precision]

    def is_applicable(self, values):
        if values is None:
            return True
        if isinstance(values, np.ndarray):
            return True
        if isinstance(values, numbers.Number):
            return True
        if isinstance(values, bool):
            return True
        if scipy.sparse.issparse(values):
            return True
        if isinstance(values, collections.Iterable):
            try:
                for value in values:
                    if not self.is_applicable(value):
                        return False
                return True
            except:
                return False
        return False

    # --- Abstract math functions ---

    def as_tensor(self, x, convert_external=True):
        if self.is_tensor(x, only_native=convert_external):
            array = x
        else:
            array = np.array(x)
        # --- Enforce Precision ---
        if not isinstance(array, numbers.Number):
            if array.dtype in (np.float16, np.float32, np.float64, np.longdouble) and self.has_fixed_precision:
                array = self.to_float(array)
        return array

    def is_tensor(self, x, only_native=False):
        if not only_native and isinstance(x, numbers.Number):
            return True
        if isinstance(x, np.ndarray):
            return x.dtype != np.object
        else:
            return False

    def copy(self, tensor, only_mutable=False):
        return np.copy(tensor)

    def equal(self, x, y):
        """ array equality comparison """
        return np.equal(x, y)

    def divide_no_nan(self, x, y):
        with np.errstate(divide='ignore', invalid='ignore'):
            result = x / y
        return np.where(y == 0, 0, result)

    def random_uniform(self, shape, low=0, high=1):
        return np.random.uniform(low, high, shape).astype(self.precision_dtype)

    def random_normal(self, shape):
        return np.random.standard_normal(shape).astype(self.precision_dtype)

    def range(self, start, limit=None, delta=1, dtype=None):
        """ range syntax to arange syntax """
        if limit is None:
            start, limit = 0, start
        return np.arange(start, limit, delta, dtype)

    def tile(self, value, multiples):
        return np.tile(value, multiples)

    def stack(self, values, axis=0):
        return np.stack(values, axis)

    def concat(self, values, axis):
        return np.concatenate(values, axis)

    def pad(self, value, pad_width, mode='constant', constant_values=0):
        passes = split_multi_mode_pad(self.ndims(value), PadSettings(pad_width, mode, constant_values), split_by_constant_value=False)
        for pad_pass in passes:
            value = self._single_mode_pad(value, *pad_pass)
        return value

    def _single_mode_pad(self, value, pad_width, single_mode, constant_values=0):
        assert single_mode in ('constant', 'symmetric', 'circular', 'reflect', 'replicate'), single_mode
        if single_mode.lower() == 'constant':
            return np.pad(value, pad_width, 'constant', constant_values=constant_values)
        else:
            if single_mode in ('circular', 'replicate'):
                single_mode = {'circular': 'wrap', 'replicate': 'edge'}[single_mode]
            return np.pad(value, pad_width, single_mode)

    def reshape(self, value, shape):
        return np.reshape(value, shape)

    def sum(self, value, axis=None, keepdims=False):
        return np.sum(value, axis=axis, keepdims=keepdims)

    def prod(self, value, axis=None):
        if not isinstance(value, np.ndarray):
            value = np.array(value)
        if value.dtype == bool:
            return np.all(value, axis=axis)
        return np.prod(value, axis=axis)

    def where(self, condition, x=None, y=None):
        if x is None or y is None:
            return np.argwhere(condition)
        return np.where(condition, x, y)

    def py_func(self, func, inputs, Tout, shape_out, stateful=True, name=None, grad=None):
        result = func(*inputs)
        assert result.dtype == Tout, "returned value has wrong type: {}, expected {}".format(result.dtype, Tout)
        assert result.shape == shape_out, "returned value has wrong shape: {}, expected {}".format(result.shape, shape_out)
        return result

    def resample(self, inputs, sample_coords, interpolation='linear', boundary='constant', constant_values=0):
        assert interpolation == 'linear'
        return general_grid_sample_nd(inputs, sample_coords, boundary, constant_values, self)

    def zeros_like(self, tensor):
        return np.zeros_like(tensor)

    def ones_like(self, tensor):
        return np.ones_like(tensor)

    def mean(self, value, axis=None, keepdims=False):
        return np.mean(value, axis, keepdims=keepdims)

    def dot(self, a, b, axes):
        return np.tensordot(a, b, axes)

    def matmul(self, A, b):
        return np.stack([A.dot(b[i]) for i in range(b.shape[0])])

    def einsum(self, equation, *tensors):
        return np.einsum(equation, *tensors)

    def while_loop(self, cond, body, loop_vars, shape_invariants=None, parallel_iterations=10, back_prop=True,
                   swap_memory=False, name=None, maximum_iterations=None):
        i = 0
        while cond(*loop_vars):
            if maximum_iterations is not None and i == maximum_iterations:
                break
            loop_vars = body(*loop_vars)
            i += 1
        return loop_vars

    def abs(self, x):
        return np.abs(x)

    def sign(self, x):
        return np.sign(x)

    def round(self, x):
        return np.round(x)

    def ceil(self, x):
        return np.ceil(x)

    def floor(self, x):
        return np.floor(x)

    def max(self, x, axis=None, keepdims=False):
        return np.max(x, axis, keepdims=keepdims)

    def min(self, x, axis=None, keepdims=False):
        return np.min(x, axis, keepdims=keepdims)

    def with_custom_gradient(self, function, inputs, gradient, input_index=0, output_index=None, name_base="custom_gradient_func"):
        return function(*inputs)

    def maximum(self, a, b):
        return np.maximum(a, b)

    def minimum(self, a, b):
        return np.minimum(a, b)

    def clip(self, x, minimum, maximum):
        return np.clip(x, minimum, maximum)

    def sqrt(self, x):
        return np.sqrt(x)

    def exp(self, x):
        return np.exp(x)

    def conv(self, tensor, kernel, padding="SAME"):
        """ apply convolution of kernel on tensor """
        assert tensor.shape[-1] == kernel.shape[-2]
        # kernel = kernel[[slice(None)] + [slice(None, None, -1)] + [slice(None)]*(len(kernel.shape)-3) + [slice(None)]]
        if padding.lower() == "same":
            result = np.zeros(tensor.shape[:-1] + (kernel.shape[-1],), dtype=self.precision_dtype)
        elif padding.lower() == "valid":
            valid = [tensor.shape[i + 1] - (kernel.shape[i] + 1) // 2 for i in range(tensor_spatial_rank(tensor))]
            result = np.zeros([tensor.shape[0]] + valid + [kernel.shape[-1]], dtype=self.precision_dtype)
        else:
            raise ValueError("Illegal padding: %s" % padding)
        for batch in range(tensor.shape[0]):
            for o in range(kernel.shape[-1]):
                for i in range(tensor.shape[-1]):
                    result[batch, ..., o] += scipy.signal.correlate(tensor[batch, ..., i], kernel[..., i, o], padding.lower())
        return result

    def expand_dims(self, a, axis=0, number=1):
        for _i in range(number):
            a = np.expand_dims(a, axis)
        return a

    def shape(self, tensor):
        return np.shape(tensor)

    def staticshape(self, tensor):
        return np.shape(tensor)

    def to_float(self, x, float64=False):
        if float64:
            warnings.warn('float64 argument is deprecated, set Backend.precision = 64 to use 64 bit operations.', DeprecationWarning)
            return np.array(x).astype(np.float64)
        else:
            return np.array(x).astype(self.precision_dtype)

    def to_int(self, x, int64=False):
        return np.array(x).astype(np.int64 if int64 else np.int32)

    def to_complex(self, x):
        x = self.as_tensor(x)
        if x.dtype in (np.complex64, np.complex128):
            return x
        elif x.dtype == np.float64:
            return x.astype(np.complex128)
        else:
            return x.astype(np.complex64)

    def cast(self, x, dtype):
        return np.array(x).astype(dtype)

    def gather(self, values, indices):
        return values[indices]

    def gather_nd(self, values, indices, batch_dims=0):
        assert indices.shape[-1] == self.ndims(values) - batch_dims - 1
        if batch_dims == 0:
            indices_list = self.unstack(indices, axis=-1)
            return values[indices_list]
        for dim in range(batch_dims):
            assert indices.shape[dim] == values.shape[dim] or values.shape[dim] == 1 or indices.shape[dim] == 1, 'Batch dimension %d does not match: %s (values) and %s (indices)' % (dim, values.shape, indices.shape)
        values_batch_max = np.array(values.shape[:batch_dims]) - 1
        indices_batch_max = np.array(indices.shape[:batch_dims]) - 1

        def inner_gather_nd(*pos):
            batch_idx_values = tuple([np.minimum(pos[i], values_batch_max[i]) for i in range(len(pos))])
            values_batch = values[batch_idx_values]
            batch_idx_indices = tuple([np.minimum(pos[i], indices_batch_max[i]) for i in range(len(pos))])
            indices_batch = indices[batch_idx_indices]
            result = values_batch[self.unstack(indices_batch, axis=-1)]
            return result
        # --- Iterate over batch dimensions ---
        batch_pos = np.meshgrid(*[range(dim) for dim in indices.shape[:batch_dims]], indexing='ij')
        batch_pos = np.stack(batch_pos, axis=-1).reshape([-1, batch_dims])
        result = np.empty(indices.shape[:-1] + (values.shape[-1],), values.dtype)
        for i in range(batch_pos.shape[0]):
            gathered = inner_gather_nd(*batch_pos[i])
            result[tuple(batch_pos[i])] = gathered
        return result

    def std(self, x, axis=None, keepdims=False):
        return np.std(x, axis, keepdims=keepdims)

    def boolean_mask(self, x, mask):
        return x[mask]

    def isfinite(self, x):
        return np.isfinite(x)

    def any(self, boolean_tensor, axis=None, keepdims=False):
        return np.any(boolean_tensor, axis=axis, keepdims=keepdims)

    def all(self, boolean_tensor, axis=None, keepdims=False):
        return np.all(boolean_tensor, axis=axis, keepdims=keepdims)

    def scatter(self, points, indices, values, shape, duplicates_handling='undefined'):
        indices = self.unstack(indices, axis=-1)
        array = np.zeros(shape, self.precision_dtype if self.has_fixed_precision else values.dtype)
        if duplicates_handling == 'add':
            np.add.at(array, tuple(indices), values)
        elif duplicates_handling == 'mean':
            count = np.zeros(shape, np.int32)
            np.add.at(array, tuple(indices), values)
            np.add.at(count, tuple(indices), 1)
            count = np.maximum(1, count)
            return array / count
        else:  # last, any, undefined
            array[indices] = values
        return array

    def fft(self, x):
        rank = len(x.shape) - 2
        assert rank >= 1
        if rank == 1:
            return np.fft.fft(x, axis=1)
        elif rank == 2:
            return np.fft.fft2(x, axes=[1, 2])
        else:
            return np.fft.fftn(x, axes=list(range(1, rank + 1)))

    def ifft(self, k):
        rank = len(k.shape) - 2
        assert rank >= 1
        if rank == 1:
            return np.fft.ifft(k, axis=1)
        elif rank == 2:
            return np.fft.ifft2(k, axes=[1, 2])
        else:
            return np.fft.ifftn(k, axes=list(range(1, rank + 1)))

    def imag(self, complex_arr):
        return np.imag(complex_arr)

    def real(self, complex_arr):
        return np.real(complex_arr)

    def sin(self, x):
        return np.sin(x)

    def cos(self, x):
        return np.cos(x)

    def dtype(self, array):
        if isinstance(array, float):
            return self.precision_dtype
        if not isinstance(array, np.ndarray):
            array = np.array(array)
        return array.dtype

    def sparse_tensor(self, indices, values, shape):
        return scipy.sparse.csc_matrix((values, self.unstack(indices, -1)), shape=shape)


def clamp(coordinates, shape):
    assert coordinates.shape[-1] == len(shape)
    for i in range(len(shape)):
        coordinates[...,i] = np.maximum(0, np.minimum(shape[i] - 1, coordinates[..., i]))
    return coordinates


def tensor_spatial_rank(field):
    dims = len(field.shape) - 2
    assert dims > 0, "channel has no spatial dimensions"
    return dims
